# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import re
from typing import List, Optional

from . import get_config


def select_texture_size(texture_sizes: List[int]) -> Optional[int]:
    TEXTURE_SIZE = get_config()["TEXTURE_SIZE"]
    expr = re.compile(r"(?P<OP>(min|max))( (?P<G>(>=|<=)) (?P<NUM>[0-9]+))?")
    match = expr.match(TEXTURE_SIZE)
    if not texture_sizes:
        return None

    assert match
    operator = match.group("OP")
    comparison = match.group("G")
    value = match.group("NUM")

    if operator and comparison and value:
        if comparison == ">=":
            filtered = [x for x in texture_sizes if x >= int(value)]
        else:
            filtered = [x for x in texture_sizes if x <= int(value)]

        if len(filtered) > 0:
            if operator == "min":
                return min(filtered)
            else:
                return max(filtered)
        else:
            # Rule tried to filter based on size, but no size fell within interval.
            # Choose a texture based on the >= or <=
            if operator == "min" and comparison == ">=":
                return max(texture_sizes)
            elif operator == "max" and comparison == "<=":
                return min(texture_sizes)
            # For the remaining two, user either set min <= (<= is redundant)
            elif operator == "min":
                return min(texture_sizes)
            # or max >= (>= is redundant)
            else:
                return max(texture_sizes)

    elif operator and comparison is None and value is None:
        if match.group("OP") == "min":
            return min(texture_sizes)
        else:
            return max(texture_sizes)
    else:
        raise Exception('Invalid TEXTURE_SIZE config: "{}"'.format(TEXTURE_SIZE))



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import json
import lzma
import os
import shutil
from logging import error, info, warning
from typing import Dict, List, Optional, Set, cast

from portmod.atom import Atom
from portmod.colour import green
from portmod.download import download_mod
from portmod.globals import env
from portmod.loader import (
    _delete_state,
    _sandbox_execute_pybuild,
    _state_path,
    load_installed_pkg,
)
from portmod.lock import vdb_lock

from .cache import clear_cache_for_path
from .fs.util import get_tree_size, onerror
from .functools import clear_install_cache
from .l10n import l10n
from .parsers.manifest import Manifest
from .perms import Permissions
from .pybuild import InstalledPybuild, Pybuild, Source
from .rebuild import get_rebuild_manifest
from .vfs import _cleanup_tmp_archive_dir


class PhaseState:
    A: List[Source]
    FILESDIR: str
    T: Optional[str]
    D: str
    USE: Set[str]
    WORKDIR: str
    ROOT: str
    UNFETCHED: List[Source]
    S: Optional[str]

    def __init__(self, build_dir: Optional[str] = None):
        if build_dir:
            self.T = os.path.join(build_dir, "temp")
        else:
            self.T = None

    @classmethod
    def from_json(cls, dictionary: Dict):
        new = PhaseState()
        for key, value in dictionary.items():
            setattr(new, key, value)
        if hasattr(new, "A"):
            new.A = [Source(**x) for x in cast(Dict, new.A)]
        if hasattr(new, "UNFETCHED"):
            new.UNFETCHED = [Source(**x) for x in cast(Dict, new.UNFETCHED)]
        return new


def get_pkg_path(mod: Pybuild) -> str:
    return os.path.join(env.prefix().PACKAGE_DIR, mod.CATEGORY, mod.PN)


def src_unpack(
    pkg: Pybuild,
    build_dir: str,
    curdir: Optional[str] = None,
    state: PhaseState = PhaseState(),
):
    permissions = Permissions(
        rw_paths=[build_dir],
        ro_paths=[env.DOWNLOAD_DIR],
        global_read=False,
        network=True,
        tmp=state.T,
    )
    _sandbox_execute_pybuild(
        pkg.FILE,
        "unpack",
        permissions,
        save_state=True,
        init=state.__dict__,
        curdir=curdir or build_dir,
    )


def src_prepare(
    pkg: Pybuild,
    build_dir: str,
    curdir: Optional[str] = None,
    state: PhaseState = PhaseState(),
):
    # Default does nothing unless pkg.PATCHES is set
    if pkg.PATCHES or "src_prepare" in pkg.FUNCTIONS:
        permissions = Permissions(
            rw_paths=[build_dir], global_read=True, network=False, tmp=state.T
        )
        _sandbox_execute_pybuild(
            pkg.FILE,
            "prepare",
            permissions,
            save_state=True,
            init=state.__dict__,
            curdir=curdir or build_dir,
        )


def src_install(
    pkg: Pybuild,
    build_dir: str,
    curdir: Optional[str] = None,
    state: PhaseState = PhaseState(),
):
    permissions = Permissions(
        rw_paths=[build_dir], global_read=True, network=False, tmp=state.T
    )
    _sandbox_execute_pybuild(
        pkg.FILE,
        "install",
        permissions,
        save_state=True,
        init=state.__dict__,
        curdir=curdir or build_dir,
    )


def pkg_postinst(
    pkg: Pybuild, final_install: str, curdir: str, state: PhaseState = PhaseState()
):
    # Default does nothing
    if "pkg_postinst" in pkg.FUNCTIONS:
        permissions = Permissions(
            rw_paths=[final_install],
            global_read=True,
            network=False,
            tmp=state.T,
        )
        _sandbox_execute_pybuild(
            pkg.FILE,
            "postinst",
            permissions,
            save_state=True,
            init=state.__dict__,
            curdir=curdir,
        )


def pkg_prerm(pkg: Pybuild, root: str, state: PhaseState = PhaseState()):
    # Default does nothing
    if "pkg_prerm" in pkg.FUNCTIONS:
        permissions = Permissions(
            rw_paths=[root], global_read=True, network=False, tmp=state.T
        )
        _sandbox_execute_pybuild(
            pkg.FILE,
            "prerm",
            permissions,
            save_state=False,
            init=state.__dict__,
            curdir=root,
        )
    if "module" in pkg.PROPERTIES:
        from .modules import iterate_pkg_modules

        for module in iterate_pkg_modules(pkg):
            module.prerm()


def remove_pkg(mod: InstalledPybuild, reinstall: bool = False):
    """
    Removes the given mod
    @param reinstall if true, don't touch the installed DB since we'll
                      need it to finish the install
    """
    # Slow imports
    import git

    print(">>> " + l10n("pkg-removing", atom=green(mod.ATOM.CPF)))

    path = get_pkg_path(mod)

    BUILD_DIR = os.path.join(env.TMP_DIR, mod.CATEGORY, mod.P)
    state = PhaseState(BUILD_DIR)
    assert state.T

    state.USE = mod.INSTALLED_USE
    os.makedirs(state.T, exist_ok=True)

    if os.path.exists(path):
        state.ROOT = path
        pkg_prerm(mod, path, state)
        del state.ROOT

        if os.path.islink(path):
            os.remove(path)
        else:
            shutil.rmtree(path, onerror=onerror)

    db_path = os.path.join(env.prefix().INSTALLED_DB, mod.CATEGORY, mod.PN)
    if os.path.exists(db_path) and not reinstall:
        # Remove and stage changes
        gitrepo = git.Repo.init(env.prefix().INSTALLED_DB)
        gitrepo.git.rm(os.path.join(mod.CATEGORY, mod.PN), r=True, f=True)
        # Clean up unstaged files (e.g. pycache)
        with vdb_lock(write=True):
            shutil.rmtree(db_path, ignore_errors=True, onerror=onerror)
            clear_cache_for_path(os.path.join(db_path, os.path.basename(mod.FILE)))

    # Remove from pybuild cache
    path = os.path.join(env.prefix().PYBUILD_INSTALLED_CACHE, mod.CATEGORY, mod.PF)
    if os.path.exists(path):
        os.remove(path)

    # Cleanup archive dir in case vfs had to extract anything
    _cleanup_tmp_archive_dir()

    print(">>> " + l10n("pkg-finished-removing", atom=green(mod.ATOM.CPF)))
    clear_install_cache()


def install_pkg(mod: Pybuild):
    # Slow imports
    import git

    print(">>> " + l10n("pkg-installing", atom=green(mod.ATOM.CPF)))
    old_curdir = os.getcwd()
    sources = download_mod(mod)
    if sources is None:
        error(">>> " + l10n("pkg-unable-to-download", atom=green(mod.ATOM.CPF)))
        return False

    BUILD_DIR = os.path.join(env.TMP_DIR, mod.CATEGORY, mod.P)
    state = PhaseState(BUILD_DIR)
    assert state.T
    state.A = [source.as_source() for source in sources]
    state.USE = mod.get_use()[0]

    # Ensure build directory is clean
    if os.path.exists(BUILD_DIR):
        shutil.rmtree(BUILD_DIR, onerror=onerror)

    state.WORKDIR = os.path.join(BUILD_DIR, "work")
    # copy files from filesdir into BUILD_DIR/files so that they are accessible
    # from within the sandbox
    FILESDIR = os.path.join(os.path.dirname(mod.FILE), "files")
    state.FILESDIR = os.path.join(BUILD_DIR, "files")
    if os.path.exists(FILESDIR):
        shutil.copytree(FILESDIR, state.FILESDIR)
    os.makedirs(state.WORKDIR, exist_ok=True)
    os.makedirs(state.T, exist_ok=True)

    info(">>> " + l10n("pkg-unpacking"))
    # Network access is allowed exclusively during src_unpack, and
    # adds additional filesystem restrictions to the sandbox
    src_unpack(mod, BUILD_DIR, state.WORKDIR, state)

    if not mod.S:
        state.S = mod.S = mod.get_default_source_basename()

    if mod.S and os.path.exists(os.path.join(state.WORKDIR, mod.S)):
        WORKDIR = os.path.join(state.WORKDIR, mod.S)
    else:
        WORKDIR = state.WORKDIR

    info(">>> " + l10n("pkg-preparing", dir=WORKDIR))

    src_prepare(mod, BUILD_DIR, WORKDIR, state)

    info(">>> " + l10n("pkg-prepared"))

    final_install_dir = os.path.join(env.prefix().PACKAGE_DIR, mod.CATEGORY)
    os.makedirs(final_install_dir, exist_ok=True)
    final_install = os.path.join(final_install_dir, mod.PN)

    state.D = os.path.join(BUILD_DIR, "image")
    os.makedirs(state.D, exist_ok=True)
    info(">>> " + l10n("pkg-installing-into", dir=state.D, atom=green(mod.ATOM.CPF)))
    src_install(mod, BUILD_DIR, WORKDIR, state)
    info(">>> " + l10n("pkg-installed-into", dir=state.D, atom=green(mod.ATOM.CPF)))

    os.chdir(env.TMP_DIR)

    if os.path.islink(state.D):
        installed_size = 0.0
    else:
        installed_size = get_tree_size(state.D) / 1024 / 1024

    build_size = get_tree_size(WORKDIR) / 1024 / 1024

    info("")
    info(f' {green("*")} ' + l10n("pkg-final-size-build", size=build_size))
    info(f' {green("*")} ' + l10n("pkg-final-size-installed", size=installed_size))
    info("")

    # If a previous version of this mod was already installed,
    # remove it before doing the final copy
    old_mod = load_installed_pkg(Atom(mod.CPN))
    db_path = os.path.join(env.prefix().INSTALLED_DB, mod.CATEGORY, mod.PN)
    if old_mod:
        remove_pkg(old_mod, os.path.exists(db_path) and mod.INSTALLED)

    info(
        ">>> "
        + l10n("pkg-installing-into", dir=final_install, atom=green(mod.ATOM.CPF))
    )

    if os.path.islink(final_install):
        os.remove(final_install)

    if os.path.exists(final_install):
        warning(l10n("pkg-existing-install-dir"))
        shutil.rmtree(final_install, onerror=onerror)

    # base/morrowind links the D directory to the morrowind install.
    # Copy the link, not the morrowind install
    if os.path.islink(state.D):
        linkto = os.readlink(state.D)
        os.symlink(linkto, final_install)
    else:
        shutil.copytree(state.D, final_install)

    state.ROOT = final_install
    pkg_postinst(mod, final_install, state.ROOT, state)

    # If installed database exists and there is no old mod, remove it
    if os.path.exists(db_path) and not old_mod:
        shutil.rmtree(db_path, onerror=onerror)

    with vdb_lock(write=True):
        # Update db entry for installed mod
        gitrepo = git.Repo.init(env.prefix().INSTALLED_DB)
        os.makedirs(db_path, exist_ok=True)

        # Copy pybuild to DB
        # unless source pybuild is in DB (i.e we're reinstalling)
        if not mod.FILE.startswith(db_path):
            shutil.copy(mod.FILE, db_path)
        gitrepo.git.add(os.path.join(mod.CATEGORY, mod.PN, os.path.basename(mod.FILE)))

        manifest_path = os.path.join(os.path.dirname(mod.FILE), "Manifest")
        if os.path.exists(manifest_path):
            # Copy Manifest to DB
            if not mod.FILE.startswith(db_path):
                shutil.copy(manifest_path, db_path)
            gitrepo.git.add(os.path.join(mod.CATEGORY, mod.PN, "Manifest"))

        def add_installed(field: str, value: str):
            with open(os.path.join(db_path, field), "w") as use:
                print(value, file=use)
            gitrepo.git.add(os.path.join(mod.CATEGORY, mod.PN, field))

        # Copy installed use configuration to DB
        add_installed("USE", " ".join(mod.get_use()[0]))
        # Copy repo pybuild was from to DB
        add_installed("REPO", mod.REPO)

        def fix_common(depstring: str):
            """
            Adds operator to dependencies in the common category
            to ensure this package is rebuilt if they change.
            """
            deps = depstring.split()
            for index, dep in enumerate(deps):
                if dep.startswith("common/"):
                    pkg = load_installed_pkg(Atom(dep))
                    # Note tilde operator. Revision bumps to
                    # common packages won't cause rebuilds.
                    if pkg:
                        deps[index] = f"~{dep}-{pkg.PV}"
            return " ".join(deps)

        # Store installed dependencies
        add_installed("RDEPEND", fix_common(mod.RDEPEND))

        # Copy pybuild environment to DB
        shutil.copy(
            os.path.join(_state_path(mod.FILE), "environment.xz"),
            os.path.join(db_path, "environment.xz"),
        )
        _delete_state(mod.FILE)
        gitrepo.git.add(os.path.join(mod.CATEGORY, mod.PN, "environment.xz"))

        path = os.path.join(db_path, "environment.xz")
        if os.path.exists(path):
            compressed_environment = lzma.LZMAFile(path)
            try:
                environment = json.load(compressed_environment)
            except EOFError as e:
                raise RuntimeError(f"Failed to read {path}") from e

            if "REBUILD_FILES" in environment and environment["REBUILD_FILES"]:
                manifest = Manifest()
                for entry in get_rebuild_manifest(environment["REBUILD_FILES"]):
                    manifest.add_entry(entry)
                manifest.write(os.path.join(db_path, "REBUILD_FILES"))

        clear_cache_for_path(os.path.join(db_path, os.path.basename(mod.FILE)))

        os.chdir(old_curdir)
        print(">>> " + l10n("pkg-installed", atom=green(mod.ATOM.CPF)))
        info("")

        if not env.DEBUG:
            shutil.rmtree(BUILD_DIR, onerror=onerror)
            # Cleanup archive dir in case vfs had to extract anything
            _cleanup_tmp_archive_dir()
            info(">>> " + l10n("cleaned-up", dir=BUILD_DIR))

        clear_install_cache()
        return True



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import re
import sys
from typing import List, Optional, Tuple

from portmod.colour import bright, lgreen, lred

from .l10n import l10n


def strtobool(value: str) -> bool:
    """
    Returns true if the value is a string with contents representing true
    and false if the value is a string with contents representing false

    True options include: yes, y, true, t, 1
    False options include: no, n, false, f, 1

    This function is localized.
    """
    _true = {
        l10n("yes-short"),
        l10n("yes").lower(),
        l10n("true").lower(),
        l10n("true-short"),
        "1",
    }
    _false = {
        l10n("no-short"),
        l10n("no").lower(),
        l10n("false").lower(),
        l10n("false-short"),
        "0",
    }

    if value.lower() in _true:
        return True
    if value.lower() in _false:
        return False
    raise ValueError(f"Invalid boolean value {value}")


def prompt_bool(question):
    sys.stdout.write(
        "{} [{}/{}]: ".format(
            question, bright(lgreen(l10n("yes"))), bright(lred(l10n("no")))
        )
    )
    while True:
        try:
            return strtobool(input().lower())
        except ValueError:
            sys.stdout.write(
                l10n(
                    "prompt-invalid-response",
                    yes=bright(lgreen(l10n("yes"))),
                    no=bright(lred(l10n("no"))),
                )
            )


def prompt_options(question: str, options: List[Tuple[str, str]]) -> str:
    print(question)
    for option, desc in options:
        print(option + ":", desc)
    sys.stdout.write("[{}]: ".format("/".join([option for option, _ in options])))
    option_set = {option for option, _ in options}
    while True:
        result = input().strip()
        if result in option_set:
            return result
        else:
            sys.stdout.write(
                l10n(
                    "prompt-invalid-response-multiple",
                    options="/".join([option for option, _ in options]),
                )
            )


def parse_num_list(string):
    if string == "":
        return list()

    m = re.match(r"(\d+)(?:-(\d+))?$", string)
    if not m:
        raise TypeError(l10n("prompt-invalid-range"))
    start = m.group(1)
    end = m.group(2) or start
    return list(range(int(start, 10), int(end, 10) + 1))


def prompt_num_multi(question, max_val):
    print("{}: ".format(question))
    while True:
        try:
            result = [y for x in input().split(",") for y in parse_num_list(x)]
            if next(filter(lambda x: x > max_val or x < 0, result), None):
                print(l10n("prompt-range-too-large-number", max=max_val))
            else:
                return result
        except ValueError:
            print(l10n("prompt-invalid-range-multi", max=max_val))


def prompt_num(question, max_val, cancel=False):
    print("{}: ".format(question))
    while True:
        try:
            result = int(input())
            if result > max_val or result < 0:
                if result == -1 and cancel:
                    return result

                print(l10n("prompt-range-too-large-number", max=max_val))
            else:
                return result
        except ValueError:
            print(l10n("prompt-invalid-range", max=max_val))


def prompt_str(question: str, default: Optional[str] = None) -> Optional[str]:
    """
    Prompts for string input

    If the user enters an EOF, returns None
    """
    if default:
        print(f'{question} (default "{default}"): ', end="")
    else:
        print(f"{question}: ", end="")
    while True:
        try:
            result = input()
        except EOFError:
            return None

        if result:
            return result

        return default



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Portmod config tests
"""

import pytest

from portmod.config import get_config
from portmod.globals import env

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_profile_only_variables(setup):
    """
    Tests that sorting the config files works properly
    """
    get_config()
    with open(env.prefix().PORTMOD_CONFIG, "w") as configfile:
        print(
            """
USE_EXPAND = "FOO"
""",
            file=configfile,
        )
    get_config.cache_clear()
    with pytest.raises(UserWarning):
        get_config()

    with open(env.prefix().PORTMOD_CONFIG, "w") as configfile:
        print(
            """
ARCH = "BAR"
""",
            file=configfile,
        )

    get_config.cache_clear()
    with pytest.raises(UserWarning):
        get_config()

    with open(env.prefix().PORTMOD_CONFIG, "w") as configfile:
        print(
            """
TEST_PROFILE_ONLY = "BAR"
""",
            file=configfile,
        )

    get_config.cache_clear()
    with pytest.raises(UserWarning):
        get_config()



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import sys
import traceback
from logging import error
from typing import Optional

from portmod.atom import Atom
from portmod.config.use import InvalidFlag, add_use, remove_use
from portmod.globals import env
from portmod.l10n import l10n
from portmod.lock import exclusive_lock


def add_use_parser(subparsers, parents):
    """
    Main function for the omwuse executable
    """

    parser = subparsers.add_parser("use", help=l10n("use-help"), parents=parents)
    parser.add_argument("-E", metavar=l10n("flag-placeholder"), help=l10n("use-enable"))
    parser.add_argument(
        "-D", metavar=l10n("flag-placeholder"), help=l10n("use-disable")
    )
    parser.add_argument("-R", metavar=l10n("flag-placeholder"), help=l10n("use-remove"))
    parser.add_argument(
        "-p", metavar=l10n("atom-placeholder"), help=l10n("use-package")
    )

    @exclusive_lock()
    def use_main(args):
        if not args.E and not args.D and not args.R and not args.p:
            parser.print_help()
            sys.exit(2)
        try:
            atom: Optional[Atom]
            if args.p is not None:
                atom = Atom(args.p)
            else:
                atom = None

            if args.E or args.D:
                if args.E:
                    add_use(args.E, atom)
                elif args.D:
                    add_use(args.D, atom, disable=True)

            if args.R:
                remove_use(args.R, atom)
        except InvalidFlag as e:
            if env.DEBUG:
                traceback.print_exc()
            error(f"{e}")
        except Exception as e:
            traceback.print_exc()
            error(f"{e}")

    parser.set_defaults(func=use_main)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
BLACKLIST and WHITELIST tests
"""

import os
import shutil
import zipfile

import pytest

from portmod.fs.util import get_hash
from portmod.globals import env
from portmod.loader import load_file
from portmod.package import install_pkg
from portmod.repo import Repo

from .env import setup_env, tear_down_env
from .test_loader import TMP_REPO, create_pybuild

LISTED = set(
    map(
        os.path.normpath,  # type:  ignore
        ["directory/foo.txt", "directory/bar.txt", "directory/bar.png"],
    )
)
OTHER = set(map(os.path.normpath, ["foo.txt", "directory/bar.gif"]))  # type:  ignore


@pytest.fixture(autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    env.REPOS.append(Repo("test", TMP_REPO))
    yield dictionary
    tear_down_env()
    shutil.rmtree(TMP_REPO)


def create_zip():
    """Creates test zip file"""
    os.chdir(env.TMP_DIR)
    os.makedirs(env.DOWNLOAD_DIR, exist_ok=True)
    get_hash.cache_clear()
    if not os.path.exists(os.path.join(env.DOWNLOAD_DIR, "test.zip")):
        with zipfile.ZipFile(os.path.join(env.DOWNLOAD_DIR, "test.zip"), "w") as myzip:
            for file in LISTED | OTHER:
                if os.path.dirname(file):
                    os.makedirs(os.path.dirname(file), exist_ok=True)
                with open(file, "w") as filep:
                    print("", file=filep)
                myzip.write(file, file)  # type:  ignore


def test_whitelist(setup):
    """
    Tests that InstallDir whitelisting works properly
    """
    pybuild = """
import os
import sys
from pybuild import Pybuild1, InstallDir

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
    SRC_URI="test.zip"

    INSTALL_DIRS=[
        InstallDir(".", WHITELIST=["directory/*.txt", "directory/*.png", "*.jpg"]),
    ]
    """
    create_zip()
    file = create_pybuild(pybuild, manifest=True)
    mod = load_file(file)
    install_pkg(mod)

    base = os.path.join(env.prefix().PACKAGE_DIR, "test", "test")
    whitelisted = LISTED.copy()
    for root, _, files in os.walk(base):
        for file in files:
            path = os.path.relpath(root, base)
            fullpath = os.path.normpath(os.path.join(path, file))
            if fullpath in whitelisted:
                whitelisted.discard(fullpath)  # type:  ignore
            else:
                raise Exception(f"File {fullpath} is not in the whitelist!")

    # Only the whitelisted file can be present
    assert not whitelisted


def test_blacklist(setup):
    """
    Tests that InstallDir blacklisting works properly
    """
    pybuild = """
import os
import sys
from pybuild import Pybuild1, InstallDir

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
    SRC_URI="test.zip"

    INSTALL_DIRS=[
        InstallDir(".", BLACKLIST=["directory/*.txt", "directory/*.png", "*.jpg"]),
    ]
    """
    create_zip()
    file = create_pybuild(pybuild, manifest=True)
    mod = load_file(file)
    install_pkg(mod)

    other_files = OTHER.copy()

    base = os.path.join(env.prefix().PACKAGE_DIR, "test", "test")
    for root, _, files in os.walk(base):
        for file in files:
            path = os.path.relpath(root, base)
            fullpath = os.path.normpath(os.path.join(path, file))
            if fullpath in LISTED:
                raise Exception(f"File {fullpath} is in the black!")
            if fullpath in other_files:
                other_files.discard(fullpath)  # type:  ignore

    assert not other_files



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests file master detection
"""

import os

import pytest

from portmod.globals import env
from portmod.masters import get_masters
from portmod.merge import configure

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup_repo():
    """sets up and tears down test environment"""
    yield setup_env("test")
    tear_down_env()


def test_masters_esp():
    """Tests that we detect esp masters properly"""

    configure(["test/quill-of-feyfolken-2.0.2"], no_confirm=True)
    path = os.path.join(
        env.prefix().PACKAGE_DIR, "test", "quill-of-feyfolken", "Quill of Feyfolken.esp"
    )
    masters = get_masters(path)
    assert len(masters) == 1
    assert "Morrowind.esm" in masters



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
import sys
from functools import lru_cache
from typing import List
from zipfile import ZipFile

from portmod.execute import execute


class UnsupportedArchive(Exception):
    """Exception raised when encountering archive types that are not supported"""


def list_archive(file: str) -> List[str]:
    """Lists the contents of the given archive"""
    _, ext = os.path.splitext(file)
    fmt_ext = ext.lstrip(".").lower()
    if fmt_ext == "zip":
        return zip_list(file)
    if fmt_ext == "bsa":
        return bsa_list(file)
    raise UnsupportedArchive(f"No handler available for extension {fmt_ext}")


def extract_archive_file(archive: str, file: str, output_dir: str):
    """Extracts the given file from the archive to the output_dir"""
    _, ext = os.path.splitext(archive)
    fmt_ext = ext.lstrip(".").lower()
    if fmt_ext == "zip":
        return zip_extract(archive, file, output_dir)
    if fmt_ext == "bsa":
        return bsa_extract(archive, file, output_dir)
    raise UnsupportedArchive(f"No handler available for extension {fmt_ext}")


def zip_list(archive: str) -> List[str]:
    """Lists the files in the given zip archive"""
    with ZipFile(archive) as file:
        return file.namelist()


def zip_extract(archive: str, file: str, output_dir: str):
    """Extracts the given file from the zip archive to the output_dir"""
    with ZipFile(archive) as zip_file:
        zip_file.extract(file, path=output_dir)


def bsa_list(archive: str) -> List[str]:
    """Lists the files in the given bsa archive"""
    return (
        (execute(f'{_bsatool_command()} list "{archive}" ', pipe_output=True) or "")
        .replace("\\", "/")
        .splitlines()
    )


def bsa_extract(archive: str, file: str, output_dir: str):
    """Extracts the given file from the bsa archive to the output_dir"""
    execute(f'{_bsatool_command()} extract "{archive}" "{file}" "{output_dir}"')


@lru_cache()
def _bsatool_command():
    if shutil.which("bsatool"):
        return "bsatool"

    if sys.platform == "win32":
        from winreg import HKEY_LOCAL_MACHINE  # pylint: disable=no-name-in-module

        from portmod.winreg import read_reg

        openmw_installs = read_reg(
            HKEY_LOCAL_MACHINE, r"Software\Wow6432Node\OpenMW.org"
        )
        if openmw_installs:
            # Add the last install found in the key. This should correspond to the newest version.
            return os.path.join(next(reversed(openmw_installs.values())), "bsatool.exe")

    raise FileNotFoundError("Could not find bsatool executable!")



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the version comparison system
"""

from portmod.atom import version_gt
from portmod.util import get_max_version


def test_simple_version():
    """Tests extremely simple version comparison"""
    v_b = "2.0"
    v_a = "1.0"
    assert get_max_version([v_a, v_b]) == v_b


def gt(atom1, atom2):
    """Tests that atom1 is considered greater than atom2"""
    assert version_gt(atom1, atom2)
    assert not version_gt(atom2, atom1)
    assert get_max_version([atom1, atom2]) == atom1
    assert get_max_version([atom2, atom1]) == atom1


def test_suffix():
    """Tests that suffixed versions come before full versions"""
    assert (
        get_max_version(["2.0_alpha", "2.0_beta", "2.0_pre", "2.0_rc", "2.0"]) == "2.0"
    )


def test_suffix_p():
    """Tests that p suffixed versions come after full versions"""
    gt("2.0_p1", "2.0")
    gt("2.0_alpha_p1", "2.0_alpha")


def test_suffix_endings():
    """Tests that p suffixed versions with integer endings order correctly"""
    gt("2.0_p2", "2.0_p1")
    gt("2.0_alpha2", "2.0_alpha1")
    gt("2.0_beta1", "2.0_alpha2")
    gt("2.0_alpha1", "2.0_alpha")
    gt("2.0_alpha1_beta2", "2.0_alpha")
    gt("2.0_alpha", "2.0_alpha_beta2")


def test_letter():
    """
    Tests that letter versions come after non-letter versions, increase in order
    and take precedence over suffixes, but not numeric components
    """
    gt("2.0a", "2.0")
    gt("2.0b", "2.0a")
    gt("2.0b_alpha", "2.0a")
    gt("2.1a", "2.0b")


def test_revision():
    """
    Tests that revisions have the lowest precedence and increase in order
    """
    gt("2.0-r1", "2.0")
    gt("2.0-r2", "2.0-r1")
    gt("2.0a-r1", "2.0-r2")
    gt("2.0_beta-r1", "2.0_alpha-r2")


def test_different_version_lengths():
    """
    Tests that having more version components is considered a greater version if the numbers are otherwise equal
    """
    gt("2.0.1", "2.0")
    gt("2.1", "2.0.1")
    gt("2.0.1_alpha", "2.0_p1-r1")



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Optional

from portmod.globals import env
from portmod.parsers.flags import add_flag, get_flags, remove_flag

from ..config import get_config
from ..pybuild import Pybuild


def add_keyword(atom, keyword):
    """Adds keyword for the given atom. Does not modify any existing keywords."""
    keyword_file = os.path.join(
        env.prefix().PORTMOD_CONFIG_DIR, "package.accept_keywords"
    )
    add_flag(keyword_file, atom, keyword)


def remove_keyword(atom, keyword):
    keyword_file = os.path.join(
        env.prefix().PORTMOD_CONFIG_DIR, "package.accept_keywords"
    )
    remove_flag(keyword_file, atom, keyword)


def get_keywords(atom):
    keyword_file = os.path.join(
        env.prefix().PORTMOD_CONFIG_DIR, "package.accept_keywords"
    )
    ACCEPT_KEYWORDS = set(get_config()["ACCEPT_KEYWORDS"])
    return get_flags(keyword_file, atom).union(ACCEPT_KEYWORDS)


def accepts_testing(arch, keywords):
    return any([keyword == "~" + arch or keyword == arch for keyword in keywords])


def accepts(accept_keywords, keywords):
    for keyword in accept_keywords:
        if keyword.startswith("~"):
            # Accepts testing on this architecture. Valid if keywords contains either
            # testing or stable for this keyword
            if keyword in keywords or keyword[1:] in keywords:
                return True
        elif keyword == "*":
            # Accepts stable on all architectures. Valid if keywords contains a stable
            # keyword for any keyword
            if any(
                [
                    not keyword.startswith("~") and not keyword.startswith("*")
                    for keyword in keywords
                ]
            ):
                return True
        elif keyword == "~*":
            # Accepts testing on all architectures. Valid if keywords contains either
            # testing or stable for any keyword
            if any([not keyword.startswith("*") for keyword in keywords]):
                return True
        elif keyword == "**":
            # Accepts any configuration
            return True
        else:  # regular keyword
            if keyword in keywords:
                return True


def get_unstable_flag(mod: Pybuild) -> Optional[str]:
    """Returns the keyword for the user's current configuration of the given mod"""
    keywords = get_keywords(mod.ATOM)
    arch = env.prefix().ARCH
    if accepts(keywords, mod.KEYWORDS):
        return None

    if accepts_testing(arch, mod.KEYWORDS) or accepts_testing(arch, keywords):
        return "~"

    return "*"



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Module for parsing and modifying sets such as @world"""

import os
from typing import Optional, Set

from portmod.atom import Atom, atom_sat
from portmod.globals import env
from portmod.repo.profiles import get_system

BUILTIN_SETS = {
    "world",
    "selected",
    "system",
    "selected-packages",
    "selected-sets",
    "rebuild",
    "modules",
    "installed",
}


def is_selected(atom: Atom) -> bool:
    """
    Returns true if and only if a package matching the given Atom is selected

    Selected packages are either system packages, packages included in the world file,
    packages included in the world sets file, or local packages
    """
    selected = get_set("world")
    for selatom in selected:
        if atom_sat(atom, selatom):
            return True
    return False


def get_set(package_set: str, parent_dir: Optional[str] = None) -> Set[Atom]:
    """
    Returns the atoms contained in the given package set

    Builtin sets include:
        world: The combination of the system and the selected sets
        selected: All packages explicitly selected,
                  either directly or indirectly through a set
        system: Packages required by the system profile
        selected-packages: Packages explicitly selected by the user
        selected-sets: Sets explicitly selected by the user
        local-packages: Packages installed locally by the user
        rebuild: Packages which have been selected for rebuilding
                 due to changes in the system
        modules: Packages with PROPERTIES="module"

    Args:
        parent_dir

    returns:
        The atoms contained in the given package set
    """
    if package_set == "world":
        return get_set("system") | get_set("selected")
    if package_set == "selected":
        return (
            get_set("selected-packages")
            | get_set("selected-sets")
            | get_set("local-packages")
        )
    elif package_set == "system":
        return get_system()
    elif package_set == "installed":
        from portmod.loader import load_all_installed

        return {pkg.ATOM for pkg in load_all_installed()}

    if parent_dir is None:
        parent_dir = env.prefix().SET_DIR

    if package_set == "local-packages":
        local_dir = os.path.join(env.prefix().PACKAGE_DIR, "local")
        if os.path.exists(local_dir):
            return set(
                map(
                    lambda name: Atom(f"local/{name}"),
                    os.listdir(local_dir),
                )
            )
        else:
            return set()

    set_file = _get_set_path(package_set, parent_dir)
    if os.path.exists(set_file):
        with open(set_file, "r") as file:
            return {Atom(s) for s in file.read().splitlines()}
    return set()


def _get_set_path(package_set: str, parent_dir: Optional[str] = None) -> str:
    if parent_dir is None:
        parent_dir = env.prefix().SET_DIR

    if package_set == "installed":
        raise RuntimeError("The @installed set cannot be modified by this function!")
    if package_set == "selected-packages":
        return os.path.join(env.prefix().PORTMOD_LOCAL_DIR, "world")
    if package_set == "selected-sets":
        return os.path.join(env.prefix().PORTMOD_LOCAL_DIR, "world_sets")
    if package_set in ("rebuild", "modules"):
        return os.path.join(env.prefix().PORTMOD_LOCAL_DIR, package_set)
    return os.path.join(parent_dir, package_set)


def add_set(package_set: str, atom: Atom, parent_dir: Optional[str] = None):
    """
    Adds an atom to a set

    args:
        package_set: Name of the set to modify
        atom: Package atom to add to the set
        parent_dir: The directory where the set file is located.
                    If not specified, defaults to the prefix's set directory
                    (a subdirectory of the config directory).
    """
    set_file = _get_set_path(package_set, parent_dir)

    os.makedirs(env.prefix().SET_DIR, exist_ok=True)
    if os.path.exists(set_file):
        with open(set_file, "r+") as file:
            for line in file:
                if atom in line:
                    break
            else:
                print(atom, file=file)
    else:
        with open(set_file, "a+") as file:
            print(atom, file=file)


def remove_set(package_set: str, atom: Atom, parent_dir: Optional[str] = None):
    """
    Removes an atom from a set

    args:
        package_set: Name of the set to modify
        atom: Package atom to remove from the set. Must match the expected atom exactly.
        parent_dir: The directory where the set file is located.
                    If not specified, defaults to the prefix's set directory
                    (a subdirectory of the config directory).
    """
    set_file = _get_set_path(package_set, parent_dir)
    if os.path.exists(set_file):
        with open(set_file, "r+") as f:
            new_f = f.readlines()
            f.seek(0)
            for line in new_f:
                if atom not in line:
                    f.write(line)
            f.truncate()



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from logging import error

from portmod.atom import Atom
from portmod.globals import env
from portmod.l10n import l10n
from portmod.loader import load_installed_pkg


def get_packages(path: str):
    for category in os.listdir(path):
        if os.path.isdir(os.path.join(path, category)) and not category.startswith("."):
            for package in os.listdir(os.path.join(path, category)):
                if os.path.isdir(os.path.join(path, category, package)):
                    yield category, package


def validate(args):
    # Check that mods in the DB correspond to mods in the mods directory
    for category, package in get_packages(env.prefix().INSTALLED_DB):
        # Check that mod is installed
        if not os.path.exists(
            os.path.join(env.prefix().PACKAGE_DIR, category, package)
        ):
            error(l10n("in-database-not-installed", atom=f"{category}/{package}"))

        # Check that pybuild can be loaded
        if not load_installed_pkg(Atom(f"{category}/{package}")):
            error(
                l10n(
                    "in-database-could-not-load",
                    atom=Atom(f"{category}/{package}"),
                )
            )

    # Check that all mods in the mod directory are also in the DB
    for category, package in get_packages(env.prefix().PACKAGE_DIR):
        if not os.path.exists(
            os.path.join(env.prefix().INSTALLED_DB, category, package)
        ):
            error(l10n("installed-not-in-database", atom=f"{category}/{package}"))



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for interacting with game configuration files as defined by
Config objects in the profile
"""

import csv
import os
from typing import Dict, Set

from portmod.atom import Atom
from portmod.loader import load_installed_pkg


def usedep_matches_installed(atom: Atom) -> bool:
    mod = load_installed_pkg(atom.strip_use())
    if not mod:
        return False  # If override isn't installed, it won't be in the graph

    for flag in atom.USE:
        if flag.startswith("-") and flag.lstrip("-") in mod.INSTALLED_USE:
            return False  # Required flag is not set
        elif not flag.startswith("-") and flag not in mod.INSTALLED_USE:
            return False  # Required flag is not set

    return True


def read_userconfig(path: str) -> Dict[str, Set[str]]:
    userconfig = {}

    if os.path.exists(path):
        # Read user config
        with open(path, newline="") as csvfile:
            csvreader = csv.reader(csvfile, skipinitialspace=True)
            for row in csvreader:
                assert len(row) > 1
                atom = row[0].strip()
                if atom not in userconfig:
                    userconfig[atom] = set(map(lambda x: x.strip(), row[1:]))
                else:
                    userconfig[atom] |= set(map(lambda x: x.strip(), row[1:]))

    return userconfig



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Various utility functions
"""

import fnmatch
import os
import re
from collections import namedtuple
from typing import Any, Iterable, List, Optional, Pattern, Tuple, Union

from portmod.config.license import has_eula, is_license_accepted
from portmod.repo.keywords import accepts, accepts_testing, get_keywords

from .atom import version_gt
from .config import get_config
from .globals import env
from .pybuild import Pybuild

KeywordDep = namedtuple("KeywordDep", ["atom", "keyword"])
LicenseDep = namedtuple("LicenseDep", ["atom", "license", "is_eula"])


def get_keyword_dep(package: Pybuild) -> Optional[KeywordDep]:
    if not accepts(get_keywords(package.ATOM), package.KEYWORDS):
        arch = env.prefix().ARCH
        if accepts_testing(arch, package.KEYWORDS):
            return KeywordDep(package.ATOM.CP, "~" + arch)
        return KeywordDep("=" + package.ATOM.CP, "**")
    return None


def select_package(packages: Iterable[Pybuild]) -> Tuple[Pybuild, Any]:
    """
    Chooses a mod version based on keywords and accepts it if the license is accepted
    """
    if not packages:
        raise Exception("Cannot select mod from empty modlist")

    filtered = list(
        filter(lambda pkg: accepts(get_keywords(pkg.ATOM), pkg.KEYWORDS), packages)
    )

    keyword = None

    if filtered:
        mod = get_newest(filtered)
    else:
        arch = env.prefix().ARCH
        # No mods were accepted. Choose the best version and add the keyword
        # as a requirement for it to be installed
        unstable = list(
            filter(lambda mod: accepts_testing(arch, mod.KEYWORDS), packages)
        )
        if unstable:
            mod = get_newest(unstable)
            keyword = "~" + arch
        else:
            # There was no mod that would be accepted by enabling testing.
            # Try enabling unstable
            mod = get_newest(packages)
            keyword = "**"

    if not is_license_accepted(mod):
        return (mod, LicenseDep(mod.CPN, mod.LICENSE, has_eula(mod)))
    if keyword is not None:
        return (mod, KeywordDep("=" + mod.ATOM.CPF, keyword))

    return (mod, None)


def get_max_version(versions: Iterable[str]) -> Optional[str]:
    """
    Returns the largest version in the given list

    Version should be a valid version according to PMS section 3.2,
    optionally follwed by a revision

    Returns None if the version list is empty
    """
    newest = None
    for version in versions:
        if newest is None or version_gt(version, newest):
            newest = version
    return newest


def get_newest(packages: Iterable[Pybuild]) -> Pybuild:
    """
    Returns the newest mod in the given list based on version
    If there is a tie, returns the earlier mod in the list
    """
    max_ver = get_max_version([pkg.PVR for pkg in packages])
    for pkg in packages:
        if pkg.PVR == max_ver:
            return pkg
    raise Exception(
        f"Internal Error: get_max_version returned incorrect result {max_ver}"
    )


def fnmatch_list_to_re(patterns: List[str]) -> Pattern:
    def to_re(value: Union[str, Pattern]):
        """
        Converts fn-match string into a regular expression string

        Note that normpath may not work as expected with fn-match strings
        if forward-slashes are present inside bracketed ranges (e.g. [/../]).
        """
        if isinstance(value, Pattern):
            return value
        return fnmatch.translate(os.path.normpath(value))

    flags = 0
    if get_config().get("CASE_INSENSITIVE_FILES", False):
        flags = re.IGNORECASE
    return re.compile("|".join(map(to_re, patterns)), flags=flags)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for handling locking

Portmod locks the system as a whole, as concurrent changes to any part of the user's
configuration may break the system.

Any major system operation should acquire the exclusive lock before starting to read from
the system configuration, guaranteeing that no other process will attempt to modify the
system while the operation is in progress.

The VDB lock prevents installed packages from being loaded while a package is being installed.
"""

import os
import tempfile
from contextlib import ContextDecorator
from typing import Dict, Optional

from fasteners import InterProcessLock, InterProcessReaderWriterLock

from .globals import env
from .l10n import l10n


class ReentrantLock:
    def __init__(self, path: str):
        self.lock = InterProcessLock(path)
        self.level = 0

    def acquire(self) -> bool:
        if self.level == 0:
            if self.lock.acquire(timeout=0.2):
                self.level += 1
                return True
            return False
        else:
            self.level += 1
            return True

    def release(self):
        self.level -= 1
        if self.level == 0:
            self.lock.release()


class ReentrantRWLock:
    def __init__(self, path: str):
        self.lock = InterProcessReaderWriterLock(path)
        self.read_level = 0
        self.write_level = 0

    def acquire(self, write: bool = False):
        # We only initialize the bar if we can't immediately acquire the lock
        # This avoids printing the progressbar if there isn't another process running
        bar = None

        if write:
            if self.write_level == 0:
                if self.read_level:
                    self.lock.release_read_lock()
                while not self.lock.acquire_write_lock(timeout=0.2):
                    if not bar:
                        bar = _get_bar().start()
                    bar.update(status=l10n("acquiring-wirte-vdb"))

            self.write_level += 1
        else:
            if self.read_level == 0:
                while not self.lock.acquire_read_lock(timeout=0.2):
                    if not bar:
                        bar = _get_bar().start()
                    bar.update(status=l10n("acquiring-read-vdb"))

            self.read_level += 1

        if bar:
            bar.finish()

    def release(self, write: bool = False):
        if write:
            self.write_level -= 1
            self.lock.release_write_lock()
            bar = None
            if self.read_level:
                while not self.lock.acquire_read_lock(timeout=0.2):
                    if not bar:
                        bar = _get_bar().start()
                    bar.update(status=l10n("acquiring-read-vdb"))
            if bar:
                bar.finish()
        else:
            self.read_level -= 1
            if self.read_level == 0:
                self.lock.release_read_lock()


temp = tempfile.gettempdir()
LOCKS: Dict[str, ReentrantRWLock] = {}
EXCLUSIVE_LOCKS: Dict[str, ReentrantLock] = {}


def has_prefix_lock():
    if env.PREFIX_NAME in LOCKS:
        lock = LOCKS[env.PREFIX_NAME]
        return bool(lock.read_level + lock.write_level)
    return False


def has_prefix_exclusive_lock():
    if env.PREFIX_NAME in EXCLUSIVE_LOCKS:
        return bool(EXCLUSIVE_LOCKS[env.PREFIX_NAME].level)
    return False


def _clear_prefix_lock(prefix: Optional[str] = None):
    global LOCKS
    _prefix = prefix or env.PREFIX_NAME
    assert _prefix
    del LOCKS[_prefix]


def _get_bar():
    from progressbar import AnimatedMarker, ProgressBar, UnknownLength, Variable

    return ProgressBar(
        widgets=[
            Variable("status", format="{formatted_value}"),
            " ",
            AnimatedMarker(),
        ],
        max_value=UnknownLength,
    )


def get_prefix_lock(
    prefix: Optional[str] = None,
) -> ReentrantRWLock:
    global LOCKS
    _prefix = prefix or env.PREFIX_NAME
    if not _prefix:
        raise Exception("Cannot lock the system outside a prefix!")
    if _prefix in LOCKS:
        lock = LOCKS[_prefix]
    else:
        lock = LOCKS[_prefix] = ReentrantRWLock(
            os.path.join(temp, f"portmod.{env.PREFIX_NAME}.lock")
        )
    return lock


def get_prefix_exclusive_lock() -> ReentrantLock:
    global EXCLUSIVE_LOCKS
    if not env.PREFIX_NAME:
        raise Exception("Cannot lock the system outside a prefix!")
    if env.PREFIX_NAME in EXCLUSIVE_LOCKS:
        lock = EXCLUSIVE_LOCKS[env.PREFIX_NAME]
    else:
        lock = EXCLUSIVE_LOCKS[env.PREFIX_NAME] = ReentrantLock(
            os.path.join(temp, f"portmod.{env.PREFIX_NAME}.exclusive.lock")
        )
    return lock


class exclusive_lock(ContextDecorator):
    def __enter__(self):
        self.lock = get_prefix_exclusive_lock()
        self.prefix = env.PREFIX_NAME

        bar = None
        while not self.lock.acquire():
            if not bar:
                bar = _get_bar().start()
            bar.update(status=l10n("acquiring-exclusive"))
        if bar:
            bar.finish()
        return self

    def __exit__(self, *exc):
        self.lock.release()
        return False


class vdb_lock(ContextDecorator):
    def __init__(self, write: bool = False):
        self.write = write

    def __enter__(self):
        self.prefix = env.PREFIX_NAME
        lock = get_prefix_lock(self.prefix)
        if self.write:
            with exclusive_lock():
                lock.acquire(write=True)
        else:
            lock.acquire()
        return self

    def __exit__(self, *exc):
        get_prefix_lock(self.prefix).release(self.write)
        return False



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for performing bulk queries on the mod database and repositories
"""

import re
import sys
from collections.abc import Container
from typing import (
    AbstractSet,
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from .atom import Atom, FQAtom, atom_sat
from .colour import blue, bright, green, lblue, lgreen, red, yellow
from .config import get_config
from .config.use import get_use, get_use_expand, use_reduce
from .download import get_total_download_size
from .globals import env
from .l10n import l10n
from .loader import load_all, load_all_installed, load_installed_pkg, load_pkg
from .parsers.usestr import parse_usestr
from .pybuild import Pybuild
from .repo.metadata import get_package_metadata
from .util import get_newest


def get_maintainer_string(maintainers: Union[List[Any], Any]) -> str:
    def list_maintainers_to_human_strings(maintainers: List[Any]) -> str:
        """ return the list of maintainers as a human readible string """
        result = ""
        for maintainer_id in range(len(maintainers)):
            maintainer = str(maintainers[maintainer_id])
            if maintainer_id >= len(maintainers) - 1:  # the last
                result += maintainer
            elif maintainer_id >= len(maintainers) - 2:  # the second last
                result += maintainer + " and "
            else:
                result += maintainer + ", "
        return result

    if not isinstance(maintainers, list):
        maintainers = [maintainers]

    return list_maintainers_to_human_strings(maintainers)


def print_depgraph(
    mod: Pybuild, level: int, level_limit: int, seen: AbstractSet[Pybuild] = frozenset()
) -> int:
    """
    Recursively prints the dependency graph for the given mod.

    Note that use conditionals and or statements are ignored.
    This prints out all possible dependencies, not actual dependencies.
    """
    if level > level_limit:
        return level - 1

    deps = parse_usestr(mod.DEPEND + " " + mod.RDEPEND, token_class=Atom)
    max_level = level

    new_seen = set(seen)

    def print_token(token, conditional=None):
        atom = Atom(token)
        mod = get_newest(load_pkg(atom))
        enabled, _ = get_use(mod)

        def colour(flag):
            if flag.rstrip("?").lstrip("-!") in enabled:
                return bright(red(flag))
            return bright(blue(flag))

        use = map(colour, atom.USE)
        use_str = ""
        if atom.USE:
            use_str = f'[{" ".join(use)}]'

        if conditional:
            dep = f"( {conditional} ( {atom.strip_use()} ) ) "
        else:
            dep = f"({atom.strip_use()}) "

        print(" " * (level + 1) + f"-- {bright(green(mod.ATOM.CPF))} " + dep + use_str)
        if mod in new_seen:
            print(" " * level + "-- " + l10n("omit-already-displayed-tree"))
            return level

        new_seen.add(mod)
        return print_depgraph(mod, level + 1, level_limit, new_seen)

    for token in deps:
        if isinstance(token, list):
            if token[0] == "||":
                for inner_token in token[1:]:
                    max_level = max(level, print_token(inner_token))
            elif token[0].endswith("?"):
                for inner_token in token[1:]:
                    max_level = max(level, print_token(inner_token, token[0]))
            else:
                for inner_token in token:
                    max_level = max(level, print_token(inner_token))
        else:
            max_level = max(level, print_token(token))

    return max_level


def str_strip(value: str) -> str:
    return re.sub("(( +- +)|(:))", "", value)


def str_squelch_sep(value: str) -> str:
    return re.sub(r"[-_\s]+", " ", value)


def query(
    fields: Union[str, Iterable],
    value: str,
    strip: bool = False,
    squelch_sep: bool = False,
    insensitive: bool = False,
    installed: bool = False,
) -> Generator[Pybuild, None, None]:
    """
    Finds mods that contain the given value in the given field
    """

    def func(val: str) -> str:
        result = val
        if insensitive:
            result = result.lower()
        if strip:
            result = str_strip(result)
        if squelch_sep:
            result = str_squelch_sep(result)
        return result

    search = func(value)

    if not installed:
        mods = load_all()
    else:
        mods = load_all_installed()

    for mod in mods:
        if isinstance(fields, str):
            if (
                hasattr(mod, fields)
                and isinstance(getattr(mod, fields), Container)
                and search in func(getattr(mod, fields))
            ):
                yield mod
        else:
            if any(
                hasattr(mod, field) and search in func(getattr(mod, field))
                for field in fields
            ):
                yield mod


def query_depends(atom: Atom, all_mods=False) -> List[Tuple[FQAtom, str]]:
    """
    Finds mods that depend on the given atom
    """
    if all_mods:
        mods = load_all()
    else:
        mods = load_all_installed()

    depends = []
    for mod in mods:
        if not all_mods:
            enabled, disabled = get_use(mod)
            atoms = use_reduce(
                mod.RDEPEND, enabled, disabled, token_class=Atom, flat=True
            )
        else:
            atoms = use_reduce(mod.RDEPEND, token_class=Atom, matchall=True, flat=True)

        for dep_atom in atoms:
            if dep_atom != "||" and atom_sat(dep_atom, atom):
                depends.append((mod.ATOM, dep_atom))
    return depends


def get_flag_string(
    name: Optional[str],
    enabled: Iterable[str],
    disabled: Iterable[str],
    installed: Optional[AbstractSet[str]] = None,
    *,
    verbose: bool = True,
    display_minuses=True,
):
    """
    Displays flag configuration

    Enabled flags are displayed as blue
    If the installed flag list is passed, flags that differ from the
    installed set will be green
    if name is None, the name prefix will be omitted and no quotes will
    surround the flags
    """

    def disable(string: str) -> str:
        if display_minuses:
            return "-" + string
        return string

    flags = []
    for flag in sorted(enabled):
        if installed is not None and flag not in installed:
            flags.append(bright(lgreen(flag)))
        elif verbose:
            flags.append(red(bright(flag)))

    for flag in sorted(disabled):
        if installed is not None and flag in installed:
            flags.append(bright(lgreen(disable(flag))))
        elif verbose:
            if display_minuses:
                flags.append(blue(disable(flag)))
            else:
                flags.append(lblue(disable(flag)))

    inner = " ".join(flags)

    if not flags:
        return None

    if name:
        return f'{name}="{inner}"'

    return inner


def display_search_results(
    mods: Iterable[Pybuild], summarize: bool = True, file=sys.stdout
):
    """
    Prettily formats a list of mods for use in search results
    """
    groupedmods: Dict[str, List[Pybuild]] = {}
    for mod in mods:
        if groupedmods.get(mod.CPN) is None:
            groupedmods[mod.CPN] = [mod]
        else:
            groupedmods[mod.CPN].append(mod)

    sortedgroups = sorted(groupedmods.values(), key=lambda group: group[0].NAME)

    for group in sortedgroups:
        sortedmods = sorted(group, key=lambda mod: mod.PV)
        newest = get_newest(group)
        installed = load_installed_pkg(Atom(newest.CPN))
        download = get_total_download_size([newest])

        if installed is not None:
            installed_str = blue(bright(installed.PV))

            flags = {flag.lstrip("+") for flag in installed.IUSE if "_" not in flag}
            usestr = get_flag_string(
                None, installed.INSTALLED_USE & flags, flags - installed.INSTALLED_USE
            )
            texture_options = {
                size
                for mod in group
                for size in use_reduce(
                    installed.TEXTURE_SIZES, matchall=True, flat=True
                )
            }
            texture = next(
                (
                    use.lstrip("texture_size_")
                    for use in installed.INSTALLED_USE
                    if use.startswith("texture_size_")
                ),
                None,
            )
            if isinstance(texture, str):
                texture_string = get_flag_string(
                    "TEXTURE_SIZE", [texture], texture_options - {texture}
                )
            else:
                texture_string = None
            use_expand_strings = []
            for use in get_config().get("USE_EXPAND", []):
                if use in get_config().get("USE_EXPAND_HIDDEN", []):
                    continue
                enabled_expand, disabled_expand = get_use_expand(installed, use)
                if enabled_expand or disabled_expand:
                    string = get_flag_string(use, enabled_expand, disabled_expand, None)
                    use_expand_strings.append(string)

            installed_str += (
                " {"
                + " ".join(
                    list(filter(None, [usestr, texture_string] + use_expand_strings))
                )
                + "}"
            )
        else:
            installed_str = "not installed"

        # List of version numbers, prefixed by either (~) or ** depending on
        # keyword for user's arch. Followed by use flags, including use expand
        version_str = ""
        versions = []
        ARCH = env.prefix().ARCH
        for mod in sortedmods:
            if ARCH in mod.KEYWORDS:
                versions.append(green(mod.PV))
            elif "~" + ARCH in mod.KEYWORDS:
                versions.append(yellow("(~)" + mod.PV))
            else:
                versions.append(red("**" + mod.PV))
        version_str = " ".join(versions)
        flags = {
            flag.lstrip("+") for mod in group for flag in mod.IUSE if "_" not in flag
        }
        usestr = get_flag_string(None, [], flags, display_minuses=False)
        texture_options = {
            size
            for mod in group
            for size in use_reduce(mod.TEXTURE_SIZES, matchall=True, flat=True)
        }
        texture_string = get_flag_string(
            "TEXTURE_SIZE", [], texture_options, display_minuses=False
        )
        use_expand_strings = []
        for use in get_config().get("USE_EXPAND", []):
            if use in get_config().get("USE_EXPAND_HIDDEN", []):
                continue
            flags = {
                re.sub(f"^{use.lower()}_", "", flag)
                for flag in mod.IUSE_EFFECTIVE
                for mod in group
                if flag.startswith(f"{use.lower()}_")
            }
            if flags:
                string = get_flag_string(use, [], flags, None, display_minuses=False)
                use_expand_strings.append(string)

        version_str += (
            " {"
            + " ".join(
                list(filter(None, [usestr, texture_string] + use_expand_strings))
            )
            + "}"
        )

        # If there are multiple URLs, remove any formatting from the pybuild and
        # add padding
        homepage_str = "\n                 ".join(newest.HOMEPAGE.split())
        mod_metadata = get_package_metadata(mod)

        print(
            "{}  {}".format(green("*"), bright(newest.CPN)),
            "       {} {}".format(green(l10n("package-name")), mod.NAME),
            "       {} {}".format(
                green(l10n("package-available-versions")), version_str
            ),
            "       {} {}".format(
                green(l10n("package-installed-version")), installed_str
            ),
            "       {} {}".format(green(l10n("package-size-of-files")), download),
            "       {} {}".format(green(l10n("package-homepage")), homepage_str),
            "       {} {}".format(
                green(l10n("package-description")), str_squelch_sep(newest.DESC)
            ),
            "       {} {}".format(green(l10n("package-license")), newest.LICENSE),
            sep="\n",
            file=file,
        )

        if mod_metadata and mod_metadata.upstream:
            maintainers = mod_metadata.upstream.maintainer
            if maintainers:
                maintainers_display_strings = get_maintainer_string(maintainers)
                print(
                    "       {} {}".format(
                        green(l10n("package-upstream-author")),
                        maintainers_display_strings,
                    ),
                    file=file,
                )

        print(file=file)

    if summarize:
        print("\n" + l10n("packages-found", num=len(sortedgroups)), file=file)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
CLI for interacting with individual pybuild files
"""

import os
import py_compile
import sys
from glob import glob
from logging import error, info

from portmod.download import download_source
from portmod.globals import env
from portmod.l10n import l10n
from portmod.loader import _sandbox_execute_pybuild, load_file
from portmod.parsers.manifest import FileType, Manifest, ManifestEntry
from portmod.perms import Permissions
from portmod.pybuild import Pybuild, manifest_path
from portmod.repo import get_repo_root
from portmod.repos import Repo
from portmod.source import HashAlg


def create_manifest(mod: Pybuild):
    """
    Automatically downloads mod DIST files (if not already in cache)
    and creates a manifest file
    """
    pybuild_file = mod.FILE
    manifest = Manifest(manifest_path(pybuild_file))
    manifest.entries = {}

    directory = os.path.dirname(pybuild_file)

    for file in glob(os.path.join(directory, "*.pybuild")):
        thismod = load_file(file)
        sources = thismod._get_sources(matchall=True)

        # Add sources to manifest
        for source in sources:
            filename = download_source(mod, source)
            if filename is None:
                error("Unable to get shasum for unavailable file " + source.name)
                continue

            manifest.add_entry(
                ManifestEntry.from_path(
                    FileType.DIST, filename, source.name, [HashAlg.BLAKE3, HashAlg.MD5]
                )
            )

    # Write changes to manifest
    manifest.write()


def pybuild_validate(file_name):
    # Verify that pybuild is valid python
    py_compile.compile(file_name, doraise=True)

    # Verify fields of pybuild
    os.makedirs(env.PYBUILD_CACHE_DIR, exist_ok=True)
    _sandbox_execute_pybuild(
        file_name, "validate", Permissions(rw_paths=[env.PYBUILD_CACHE_DIR])
    )


def pybuild_manifest(file_name):
    if not os.path.exists(file_name):
        raise FileNotFoundError(l10n("file-does-not-exist", file=file_name))

    repo_root = get_repo_root(file_name)

    if repo_root is None:
        raise FileNotFoundError(l10n("repository-does-not-exist"))

    # Register repo in case it's not already in repos.cfg
    REAL_ROOT = os.path.realpath(repo_root)
    if not any([REAL_ROOT == os.path.realpath(repo.location) for repo in env.REPOS]):
        sys.path.append(os.path.join(repo_root))
        env.REPOS.append(Repo(os.path.basename(repo_root), repo_root))

    if env.ALLOW_LOAD_ERROR:
        raise Exception("Cannot allow load errors when generating manifest!")

    mod = load_file(file_name)

    create_manifest(mod)
    info(l10n("created-manifest", atom=mod.ATOM))



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import sys

import pytest

from portmod.atom import Atom, version_gt
from portmod.deps import resolve
from portmod.merge import configure
from portmod.transactions import Downgrade, New, Reinstall, Update

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_simple(setup):
    """Tests that simple dependency resolution works"""
    selected = {Atom("test/test")}
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert transactions.pkgs[0].pkg.CPN == "test/test"
    assert isinstance(transactions.pkgs[0], New)

    configure(["test/test"], no_confirm=True)
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert transactions.pkgs[0].pkg.CPN == "test/test"
    assert isinstance(transactions.pkgs[0], Reinstall)


def test_rebuild(setup):
    """
    Tests that packages are selected to be rebuilt, even if we don't
    use the Category-PackageName format
    """
    selected = {Atom("~test/test-1.0")}
    configure(selected, no_confirm=True)
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert transactions.pkgs[0].pkg.CPN == "test/test"
    assert transactions.pkgs[0].pkg.REPO == "test"
    assert isinstance(transactions.pkgs[0], Reinstall)


def test_upgrade(setup):
    """Tests that upgrades resolve correctly"""
    selected = {Atom("test/test")}
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert version_gt(transactions.pkgs[0].pkg.PVR, "1.0")
    assert isinstance(transactions.pkgs[0], Update)


def test_oneshot(setup):
    """Tests that oneshot resolves correctly"""
    selected = {Atom("test/test")}
    configure(selected, no_confirm=True)
    transactions = resolve(selected, set(), selected, set(), set())
    assert len(transactions.pkgs) == 1
    assert not version_gt(transactions.pkgs[0].pkg.PVR, "2.0")
    assert not version_gt("2.0", transactions.pkgs[0].pkg.PVR)
    assert isinstance(transactions.pkgs[0], Reinstall)
    assert not transactions.new_selected


@pytest.mark.xfail(
    sys.platform == "win32" and "APPVEYOR" in os.environ,
    reason="For some reason a file is being accessed by another process",
    raises=PermissionError,
)
def test_downgrade(setup):
    """Tests that downgrades resolve correctly"""
    configure(["=test/test-2.0"], no_confirm=True)
    selected = {Atom("=test/test-1.0")}
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert version_gt("2.0", transactions.pkgs[0].pkg.PVR)
    assert isinstance(transactions.pkgs[0], Downgrade)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests some otherwise untested parts of the interface
"""

import os
import sys

import pytest

from portmod._cli.main import main
from portmod.atom import Atom
from portmod.globals import env
from portmod.loader import load_installed_pkg

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    yield setup_env("test")
    tear_down_env()


def test_validate():
    """Tests that validate works correctly"""
    sys.argv = ["portmod", "test", "merge", "test", "test2", "--no-confirm"]
    main()
    sys.argv = ["portmod", "test", "validate"]
    main()


def test_sync():
    """Tests that portmod sync works correctly"""
    sys.argv = ["portmod", "sync"]
    main()


def test_info():
    """Tests that portmod sync works correctly"""
    sys.argv = ["portmod", "test", "info"]
    with pytest.raises(SystemExit):
        main()


def test_search():
    """Tests that portmod sync works correctly"""
    sys.argv = ["portmod", "test", "search", "test"]
    main()


def test_use():
    """Tests that portmod sync works correctly"""
    sys.argv = ["portmod", "test", "use", "-E", "foo"]
    main()
    sys.argv = ["portmod", "test", "merge", "test4", "--no-confirm"]
    main()
    pkg = load_installed_pkg(Atom("test/test4"))
    assert pkg
    assert "foo" in pkg.get_enabled_use()


@pytest.mark.xfail(
    sys.platform == "win32" and "APPVEYOR" in os.environ,
    reason="For some reason the vdb git repo can't be removed during cleanup",
    raises=PermissionError,
)
def test_mirror():
    """Tests that portmod --version works correctly"""
    sys.argv = ["portmod", "mirror", os.path.join(env.TMP_DIR, "mirror")]
    main()



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Module for loading pybuilds within a sandboxed environment"""

import json
import logging
import os
import re
import shutil
import subprocess
import sys
import traceback
from functools import lru_cache, wraps
from logging import debug, warning
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    TypeVar,
    cast,
)

import portmod
from portmod.config import get_config
from portmod.functools import prefix_aware_cache

from .atom import Atom, FQAtom, atom_sat
from .cache import (
    PreviouslyEncounteredException,
    __load_mod_from_dict_cache,
    _load_cache_dict,
    pybuild_dumper,
)
from .colour import green
from .execute import _pybuild_exec_paths, execute, sandbox_execute
from .globals import env
from .l10n import l10n
from .perms import Permissions
from .pybuild import File, InstallDir, InstalledPybuild, Pybuild
from .repo.loader import (
    _iterate_installed,
    _iterate_pybuilds,
    find_installed_path,
    get_atom_from_path,
)


class AmbiguousAtom(Exception):
    """Indicates that multiple packages from different categories match"""

    def __init__(self, atom: Atom, packages: Iterable[Atom], fq: bool = False):
        message_id = "ambiguous-atom-fq" if fq else "ambiguous-atom"
        super().__init__(
            l10n(message_id, atom=green(atom))
            + "\n"
            + green("\n  ".join(sorted(packages)))
        )


class SandboxedError(Exception):
    """Error raised when a sandboxed command fails"""


class LoadFromSandboxError(Exception):
    """Exception raised when trying to load a (uncached) package from within the sandbox"""


@lru_cache()
def _state_path(file) -> str:
    atom = get_atom_from_path(file)
    return os.path.join(env.TMP_DIR, atom.C, atom.P, "state")


def _delete_state(file):
    if os.path.exists(_state_path(file)):
        shutil.rmtree(_state_path(file))


def get_wrapper_code():
    # Preserve virtualenv
    venv_activate = ""
    if "VIRTUAL_ENV" in os.environ:
        if sys.platform == "win32":
            file = os.path.join(
                os.environ["VIRTUAL_ENV"], "scripts", "activate_this.py"
            ).replace("\\", "\\\\")
        else:
            file = os.path.join(os.environ["VIRTUAL_ENV"], "bin", "activate_this.py")
        venv_activate = (
            'exec(open("' + file + '").read(), {"__file__": "' + file + '"})'
        )

    return f"""
import sys
from os import path as osp

{venv_activate}
if __name__ == "__main__":
    # Ignore the -c argument, and this code, which are the first two arguments passed to python
    del sys.argv[0:1]
    # Third argument should be a file within the portmod module.
    # This also takes the place of the program name in the argument list so that argparse
    # handles the remaining arguments correctly
    if osp.isfile(
        osp.join(
            osp.dirname(osp.dirname(osp.realpath(sys.argv[0]))), ".portmod_not_installed"
        )
    ):
        sys.path.insert(0, osp.dirname(osp.dirname(osp.realpath(sys.argv[0]))))

    from portmod.wrapper import main

    main()
"""


def load_module(file: str, pkg: Pybuild, state):
    from portmod.modules import Module, ModuleFunction

    perms = Permissions(rw_paths=[state.CACHE], global_read=True, tmp=state.TEMP)

    module_data = _sandbox_execute_module(file, "load", permissions=perms)

    def get_func_wrapper(function_name: str):

        if function_name == "update":

            def func_wrapper(state):
                return _sandbox_execute_module(
                    file,
                    "execute",
                    function=function_name,
                    init=state.__dict__,
                    permissions=perms,
                )

        else:

            def func_wrapper(state, args):  # type: ignore
                return _sandbox_execute_module(
                    file,
                    "execute",
                    function=function_name,
                    args=args,
                    init=state.__dict__,
                    permissions=perms,
                )

        return func_wrapper

    functions = []
    for function, data in module_data.get("functions", []).items():
        functions.append(
            ModuleFunction(
                function,
                data.get("desc"),
                get_func_wrapper(function),
                data.get("options"),
                data.get("parameters"),
                state,
            )
        )

    return Module(
        module_data.get("name"),
        module_data.get("desc"),
        sorted(functions, key=lambda x: x.name),
        state,
    )


@lru_cache()
def _get_library_dirs(path: str) -> Set[str]:
    """
    Returns the directories containing the libraries
    used by the executable at the given path

    Used to attempt to detect non-standard library directories
    """
    # Note: while we could just use direct paths of the libraries, this gives a little more
    # Flexibility for providing other libraries in the sandbox
    paths = set()

    if sys.platform == "win32":
        pass
    else:
        try:
            if sys.platform == "darwin":
                lines = execute(f"otool -L {path}", pipe_output=True)
            else:
                lines = execute(f"ldd {path}", pipe_output=True)
        except subprocess.CalledProcessError:
            return set()

        # Add anything that looks like an absolute path
        for token in (lines or "").split():
            directory = None

            if token.startswith("/"):
                directory = token
            elif token.startswith("@executable_path/"):
                directory = os.path.normpath(
                    os.path.join(path, token.replace("@executable_path/", ""))
                )

            if directory:
                paths.add(os.path.dirname(directory))
    return paths


def _sandbox_execute(
    file_type: str,
    file: str,
    action: str,
    permissions: Permissions,
    *,
    save_state: bool = False,
    init: Optional[Dict[str, Any]] = None,
    curdir: Optional[str] = None,
    installed: bool = False,
    args: Optional[Any] = None,
    function: Optional[str] = None,
):
    assert file_type in ("pybuild", "module")
    python = os.path.realpath(sys.executable)
    print_actions = [
        "load",
        "cache",
        "dump-environment",
    ]

    abspath = os.path.abspath(file)
    old_curdir = os.getcwd()
    if not os.path.exists(curdir or env.TMP_DIR):
        os.makedirs(curdir or env.TMP_DIR)
    os.chdir(curdir or env.TMP_DIR)
    ro_paths = set(permissions.ro_paths)
    ro_paths.add(env.CONFIG_DIR)
    for repo in env.REPOS:
        ro_paths.add(repo.location)
    # Python site packages directories, etc.
    ro_paths |= set(sys.path)
    if sys.platform == "win32":
        ro_paths |= set(os.environ["PATH"].split(";"))
    else:
        ro_paths |= set(os.environ["PATH"].split(":"))

    if not curdir:
        ro_paths.add(env.TMP_DIR)
    ro_paths.add(os.path.dirname(abspath))
    # Detect Libraries used by executables in case of non-standard library locations
    for executable in ["python", "git", "bsatool"]:
        path = shutil.which(executable)
        if path:
            ro_paths |= _get_library_dirs(os.path.realpath(path))
    env_pickle = env.serialize().hex()
    command = [
        python,
        "-c",
        get_wrapper_code(),
        portmod.__file__,
        "--verbosity",
        logging.getLevelName(logging.root.level),
        "--env",
        env_pickle,
        file_type,
        abspath,
        action,
    ]
    rw_paths = set(permissions.rw_paths)
    rw_paths.add(env.PYBUILD_TMP_DIR)
    if save_state:
        command += ["--state-dir", _state_path(file)]
        os.makedirs(_state_path(file), exist_ok=True)
        rw_paths.add(_state_path(file))
    if init:
        # Note: quotes must be escaped
        state_string = json.dumps(init, default=pybuild_dumper)
        command += ["--initial-state", state_string]
    if args:
        command += ["--args", json.dumps(args)]
    if function:
        command += ["--module-func", function]
    try:
        result = sandbox_execute(
            command,
            Permissions(
                permissions,
                ro_paths=sorted(ro_paths),
                rw_paths=sorted(rw_paths),
            ),
            pipe_output=action in print_actions,
            exec_paths=(action not in ("cache", "validate") and _pybuild_exec_paths())
            or [],
        )
    except subprocess.CalledProcessError as err:
        debug(err)
        raise SandboxedError(
            l10n("command-failed", path=abspath, command=action)
        ) from None
    finally:
        os.chdir(old_curdir)
    return result


def _sandbox_execute_module(
    file: str,
    action: str,
    *,
    permissions: Permissions = Permissions(),
    args: Optional[Any] = None,
    function: Optional[str] = None,
    init: Optional[Dict[str, Any]] = None,
):
    """
    Modules, as they are only executed after installation, have greater permissions than
    pybuilds. They have read-only access to the entire filesystem, though not to the
    network by default.

    Write access should be done using the CONFIG_PROTECT_DIR. There is a create_file
    function which can be used to create files in the CONFIG_PROTECT_DIR that shadow
    another file in the filesystem. The user will be prompted to overwrite the shadowed
    file when the module is finished executing.
    """
    # TODO: Allow modules to request network permissions
    os.makedirs(env.prefix().CONFIG_PROTECT_DIR, exist_ok=True)
    result = _sandbox_execute(
        "module",
        file,
        action,
        Permissions(
            permissions,
            rw_paths=[env.prefix().CONFIG_PROTECT_DIR],
        ),
        init=init,
        function=function,
        args=args,
    )

    if action == "load":
        if not result:
            raise SandboxedError(f"Loading module {file} did not produce output!")
        return json.loads(result)


def _sandbox_execute_pybuild(
    file: str,
    action: str,
    permissions: Permissions,
    *,
    save_state: bool = False,
    init: Optional[Dict[str, Any]] = None,
    curdir: Optional[str] = None,
    installed: bool = False,
):
    result = _sandbox_execute(
        "pybuild",
        file,
        action,
        permissions,
        save_state=save_state,
        init=init,
        curdir=curdir,
        installed=installed,
    )
    atom = get_atom_from_path(file)

    if action == "can-update-live":
        return result is True

    if action == "load":
        pkg = json.loads(result)
        return _load_cache_dict(file, FQAtom(atom), pkg, installed=False)
    if action in ["cache", "dump-environment"]:
        return result


V = TypeVar("V", bound=Any)


def __safe_load(user_function: Callable[..., V]) -> Callable[..., Optional[V]]:
    """
    Decorator that makes a function return None if it would otherwise raise an exception
    """

    @wraps(user_function)
    def decorating_function(name, *args, **kwargs):
        try:
            return user_function(name, *args, **kwargs)
        except LoadFromSandboxError:
            return None
        except PreviouslyEncounteredException as e:
            if env.ALLOW_LOAD_ERROR:
                return None
            raise e.previous
        except Exception as e:
            warning(e)
            if env.DEBUG:
                traceback.print_exc()
            warning(l10n("could-not-load-pybuild", file=name))
            if env.ALLOW_LOAD_ERROR:
                return None
            raise e

    return cast(Callable[..., Optional[V]], decorating_function)


def load_installed_pkg(atom: Atom) -> Optional[InstalledPybuild]:
    """Loads packages from the installed database"""
    path = find_installed_path(atom)

    if path is not None:
        pkg = cast(Optional[InstalledPybuild], safe_load_file(path, installed=True))
        if pkg and atom_sat(pkg.ATOM, atom, ignore_name=True):
            return pkg

    if not atom.C or atom.C == "local":
        local_path = os.path.join(env.prefix().PACKAGE_DIR, "local", atom.PN)
        if os.path.exists(local_path):
            return _load_local_pkg(local_path)

    return None


def load_pkg_fq(atom: FQAtom) -> Pybuild:
    """
    Loads package matching fully qualified atom.

    except:
        FileNotFoundError: If the package cannot be found
        AmbiguousAtom: If multiple packages match the given atom
    """
    if atom.R.endswith("::installed") or atom.R == "installed":
        installed = load_installed_pkg(atom)
        if installed:
            return installed

        raise FileNotFoundError(l10n("not-found", atom=atom))

    packages: List[Pybuild] = []
    for file in _iterate_pybuilds(atom, atom.R):
        pkg = safe_load_file(file)
        if pkg is None:
            continue

        packages.append(pkg)

    if len(packages) > 1:
        raise AmbiguousAtom(atom, [pkg.ATOM for pkg in packages], fq=True)
    if len(packages) == 1:
        return packages[0]

    raise FileNotFoundError(l10n("not-found", atom=atom))


def load_pkg(atom: Atom, *, repo_name: Optional[str] = None) -> List[Pybuild]:
    """
    Loads all mods matching the given atom
    There may be multiple versions in different repos,
    as well versions with different version or release numbers

    :param atom: Mod atom to load.
    :param repo_name: If present, the name of the repository tree to search.
                      The masters of the given repository will also be searched.
    """
    mods = []

    for file in _iterate_pybuilds(atom, repo_name):
        mod = safe_load_file(file)

        if mod is None:
            continue

        mods.append(mod)

    if repo_name is None and env.PREFIX_NAME:
        installed = load_installed_pkg(atom)
        # Ignore the name, in case it was moved
        if installed and atom_sat(installed.ATOM, atom, ignore_name=True):
            mods.append(installed)

    return mods


def load_all() -> Generator[Pybuild, None, None]:
    for file in _iterate_pybuilds():
        mod = safe_load_file(file)
        if mod is None:
            continue

        yield mod


@prefix_aware_cache
def _load_local_pkg(package_path: str) -> InstalledPybuild:
    name = os.path.basename(package_path)
    # Use config to auto-detect special files such as plugins
    install_dir = InstallDir(".")

    def add_files(file_type, pattern, base_dir):
        if not pattern and os.path.exists(base_dir):
            getattr(install_dir, file_type).append(
                File(os.path.relpath(base_dir, package_path))
            )
        else:
            component, _, pattern = pattern.partition("/")
            for path in os.listdir(base_dir):
                if re.match(component, path):
                    add_files(file_type, pattern, os.path.join(base_dir, path))

    for file_type, pattern in get_config().get("LOCAL_FILES", {}).items():
        setattr(install_dir, file_type, [])
        add_files(file_type, pattern, package_path)

    return InstalledPybuild(
        FQAtom(f"local/{name}-0::installed"),
        INSTALL_DIRS=[install_dir],
        FILE=package_path,
        PROPERTIES="local",
        REPO="",
    )


def load_all_installed() -> Generator[InstalledPybuild, None, None]:
    """
    Returns a flat set of all installed packages
    """
    for path in _iterate_installed():
        mod = cast(Optional[InstalledPybuild], safe_load_file(path, installed=True))
        if mod:
            yield mod

    local_dir = os.path.join(env.prefix().PACKAGE_DIR, "local")
    if os.path.exists(local_dir):
        for subdir in os.listdir(local_dir):
            path = os.path.join(env.prefix().PACKAGE_DIR, "local", subdir)
            if os.path.isdir(path):
                yield _load_local_pkg(path)


def load_all_installed_map() -> Dict[str, List[InstalledPybuild]]:
    """
    Returns every single installed mod in the form of a map from their simple mod name
    to their mod object
    """
    mods: Dict[str, List[InstalledPybuild]] = {}
    for mod in load_all_installed():
        if mods.get(mod.PN) is None:
            mods[mod.PN] = [mod]
        else:
            mods[mod.PN].append(mod)
    return mods


def load_file(path: str, installed: bool = False) -> Pybuild:
    """Loads the pybuild at the given path"""
    return __load_mod_from_dict_cache(path, installed=installed)


@__safe_load
def safe_load_file(path: str, installed: bool = False) -> Optional[Pybuild]:
    """
    Loads the pybuild at the given path

    :returns: The pybuild, or None if it could not be loaded
    """
    return load_file(path, installed)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Functions for interacting with the OpenMW VFS
"""

import os
import shutil
from logging import error, info, warning
from tempfile import gettempdir
from typing import Dict, List, Set, Tuple

from .atom import Atom, atom_sat
from .config.userconf import read_userconfig, usedep_matches_installed
from .fs.archives import extract_archive_file, list_archive
from .fs.util import ci_exists
from .functools import clear_vfs_cache, vfs_cache
from .globals import env
from .l10n import l10n
from .loader import load_all_installed_map
from .parsers.list import read_list, write_list
from .parsers.usestr import use_reduce
from .pybuild import InstalledPybuild
from .tsort import CycleException, tsort


def sort_vfs_if_needed(user_function):
    """
    Decorator that sorts the vfs before executing the given function
    if it is necessary
    """

    def decorating_function(*args, **kwargs):
        if vfs_needs_sorting():
            try:
                sort_vfs()
                clear_vfs_sort()
            except CycleException as err:
                error(f"{err}")
        return user_function(*args, **kwargs)

    return decorating_function


@vfs_cache
def find_file(name: str) -> str:
    """
    Locates the path of a file within the OpenMW virtual file system

    args:
        name: The relative path within the VFS to search for

    returns:
        The absolute path of the file
    """
    # FIXME: This should respect the CASE_INSENSITIVE_FILES setting
    for directory in reversed(get_vfs_dirs()):
        path = ci_exists(os.path.join(directory, name))
        if path:
            return path

    for archive in reversed(get_vfs_archives()):
        contents = list_archive(archive)
        for file in contents:
            if os.path.normpath(file).lower() == os.path.normpath(name).lower():
                return extract_archive_file_to_tmp(archive, file)

    raise FileNotFoundError(name)


@vfs_cache
def list_dir(name: str) -> List[str]:
    """
    Locates all path of files matching the given pattern within the OpenMW
    virtual file system

    args:
        name: The relative path of the directory within the VFS
    returns:
        A list of files contained within the directory
    """
    files: Dict[str, str] = {}
    normalized = os.path.normpath(name).lower()

    # FIXME: This should respect the CASE_INSENSITIVE_FILES setting

    for directory in reversed(get_vfs_dirs()):
        path = ci_exists(os.path.join(directory, normalized))
        if path:
            for file in os.listdir(path):
                if file.lower() not in files:
                    files[file.lower()] = file

    for archive in reversed(get_vfs_archives()):
        contents = list_archive(archive)
        for file in contents:
            if (
                os.path.commonpath([normalized, os.path.normpath(file).lower()])
                == normalized
            ):
                suffix = os.path.relpath(os.path.normpath(file).lower(), normalized)
                component, _, _ = suffix.partition(os.sep)
                files[component] = component

    return sorted(files.values())


def _cleanup_tmp_archive_dir():
    path = os.path.join(gettempdir(), ".archive_files")
    if os.path.exists(path):
        shutil.rmtree(path)


def extract_archive_file_to_tmp(archive: str, file: str) -> str:
    """Extracts the given file from the archive and places it in a temprorary directory"""
    temp = gettempdir()
    output_dir = os.path.join(
        temp, ".archive_files", os.path.basename(archive), os.path.dirname(file)
    )
    os.makedirs(output_dir, exist_ok=True)
    result_file = os.path.join(output_dir, os.path.basename(file))
    extract_archive_file(archive, file, output_dir)
    if not os.path.exists(result_file):
        raise Exception(l10n("archive-extraction-failed", file=file, dest=result_file))
    return result_file


@vfs_cache
def get_vfs_dirs() -> List[str]:
    """Returns an ordered list of the VFS directories, in reverse order of priority"""
    return read_list(os.path.join(env.prefix().PACKAGE_DIR, "vfs"))


def __set_vfs_dirs__(dirs: List[str]):
    """Updates the vfs directories"""
    write_list(os.path.join(env.prefix().PACKAGE_DIR, "vfs"), dirs)


@vfs_cache
def get_vfs_archives() -> List[str]:
    """Returns an ordered list of the VFS directories, in reverse order of priority"""
    return read_list(os.path.join(env.prefix().PACKAGE_DIR, "vfs-archives"))


def __set_vfs_archives(archives: List[str]):
    """Updates the vfs directories"""
    write_list(os.path.join(env.prefix().PACKAGE_DIR, "vfs-archives"), archives)


def require_vfs_sort():
    """
    Creates a file that indicates the vfs still needs to be sorted
    """
    open(
        os.path.join(env.prefix().PORTMOD_LOCAL_DIR, ".vfs_sorting_incomplete"), "a"
    ).close()


def clear_vfs_sort():
    """Clears the file indicating the config needs sorting"""
    path = os.path.join(env.prefix().PORTMOD_LOCAL_DIR, ".vfs_sorting_incomplete")
    if os.path.exists(path):
        os.remove(path)


def vfs_needs_sorting():
    """Returns true if changes have been made since the config was sorted"""
    return os.path.exists(
        os.path.join(env.prefix().PORTMOD_LOCAL_DIR, ".vfs_sorting_incomplete")
    )


def sort_vfs():
    """Regenerates the vfs list"""
    info(l10n("sorting-vfs"))
    _sort_vfs_dirs()
    _sort_vfs_archives()
    clear_vfs_cache()


def load_userconfig(typ: str, installed_dict: Dict[str, List[InstalledPybuild]]):
    """Checks entries in userconfig and warns on errors"""
    # Keys refer to master atoms (overridden).
    # values are a set of overriding mod atomso
    user_config_path = os.path.join(
        env.prefix().PORTMOD_CONFIG_DIR, "config", f"{typ}.csv"
    )
    userconfig: Dict[str, Set[str]] = read_userconfig(user_config_path)

    for entry in userconfig.keys() | {
        item for group in userconfig.values() for item in group
    }:
        possible_mods = installed_dict.get(Atom(entry).PN, [])
        if not possible_mods:
            warning(
                l10n("user-config-not-installed", entry=entry, path=user_config_path)
            )
        elif len(possible_mods) > 1:
            warning(
                l10n(
                    "user-config-ambiguous",
                    entry=entry,
                    path=user_config_path,
                    packages=" ".join([mod.ATOM.CPF for mod in possible_mods]),
                )
            )
    return userconfig


def _sort_vfs_archives():
    installed_dict = load_all_installed_map()
    installed = [mod for group in installed_dict.values() for mod in group]

    graph: Dict[str, Set[str]] = {}
    priorities = {}

    for mod in installed:
        for install, file in mod.get_files("ARCHIVES"):
            path = mod.get_file_path(install, file)
            graph[path] = set()
            priorities[path] = mod.TIER

    userconfig = load_userconfig("archives", installed_dict)

    # Add edges in the graph for each data override
    for mod in installed:
        for install, file in mod.get_files("ARCHIVES"):
            path = mod.get_file_path(install, file)

            masters = set()
            if isinstance(file.OVERRIDES, str):
                masters |= set(use_reduce(file.OVERRIDES, mod.INSTALLED_USE))
            else:
                masters |= set(file.OVERRIDES)

            if file.NAME in userconfig:
                masters |= set(userconfig[path])

            for master in masters:
                if master in graph:
                    graph[master].add(path)
    try:
        sorted_archives = tsort(graph, priorities)
    except CycleException as error:
        raise CycleException(l10n("vfs-cycle-error"), error.cycle) from error

    __set_vfs_archives(sorted_archives)


def _sort_vfs_dirs():
    installed_dict = load_all_installed_map()
    installed = [mod for group in installed_dict.values() for mod in group]

    graph: Dict[Tuple[str, str, bool], Set[Tuple[str, str, bool]]] = {}
    priorities = {}

    userconfig = load_userconfig("install", installed_dict)

    # Determine all Directories that are enabled
    for mod in installed:
        for install in mod.get_directories():
            default = os.path.normpath(install.PATCHDIR) == "."
            path = mod.get_dir_path(install)
            graph[(mod.ATOM.CP, path, default)] = set()
            priorities[(mod.CP, path, default)] = mod.TIER

    # Add edges in the graph for each data override
    for mod in installed:
        for install in mod.get_directories():
            idefault = os.path.normpath(install.PATCHDIR) == "."
            ipath = mod.get_dir_path(install)
            parents = set(
                use_reduce(
                    mod.DATA_OVERRIDES + " " + install.DATA_OVERRIDES,
                    mod.INSTALLED_USE,
                    flat=True,
                    token_class=Atom,
                )
            ) | {
                Atom(override)
                for name in userconfig
                for override in userconfig[name]
                if atom_sat(mod.ATOM, Atom(name))
            }

            for parent in parents:
                if not usedep_matches_installed(parent):
                    continue

                for (atom, path, default) in graph:
                    if atom_sat(Atom(atom), parent) and default:
                        if Atom(atom).BLOCK:
                            # Blockers have reversed edges
                            graph[(mod.ATOM.CP, ipath, idefault)].add(
                                (atom, path, default)
                            )
                        else:
                            graph[(atom, path, default)].add((mod.CP, ipath, idefault))
    try:
        sorted_mods = tsort(graph, priorities)
    except CycleException as error:
        raise CycleException(l10n("vfs-cycle-error"), error.cycle) from error

    new_dirs = [path for _, path, _ in sorted_mods]
    __set_vfs_dirs__(new_dirs)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for interacting with the user's profile
and iterating over the various associated directories
"""

import os
from pathlib import Path
from typing import List, Set

from portmod.atom import Atom
from portmod.functools import prefix_aware_cache
from portmod.globals import env
from portmod.parsers.list import read_list


def profile_exists() -> bool:
    return bool(
        env.PREFIX_NAME
        and os.path.exists(os.path.join(env.prefix().PORTMOD_CONFIG_DIR, "profile"))
    )


def get_profile_path() -> str:
    """Returns the path to the profile directory"""
    profilepath = os.path.join(env.prefix().PORTMOD_CONFIG_DIR, "profile")
    if not os.path.exists(profilepath) or not os.path.islink(profilepath):
        raise Exception(
            f"{profilepath} does not exist.\n"
            "Please choose a profile before attempting to install packages"
        )
    # Note: Path must be resolved with pathlib to ensure that
    # the path doesn't include \\?\ on Windows
    return str(Path(os.readlink(profilepath)).resolve())


@prefix_aware_cache
def profile_parents() -> List[str]:
    """
    Produces the paths of all the parent directories for the selected profile, in order
    """
    if not env.PREFIX_NAME:
        return []

    first = get_profile_path()

    def get_parents(directory: str) -> List[str]:
        parentfile = os.path.join(directory, "parent")
        parents = []
        if os.path.exists(parentfile):
            for parent in read_list(parentfile):
                parentpath = os.path.normpath(os.path.join(directory, parent))
                parents.extend(get_parents(parentpath))
                parents.append(parentpath)

        return parents

    parents = [first]
    parents.extend(get_parents(first))

    userpath = os.path.join(env.prefix().PORTMOD_CONFIG_DIR, "profile.user")
    if os.path.exists(userpath):
        parents.append(userpath)
        parents.extend(get_parents(userpath))

    return parents


def get_system() -> Set[Atom]:
    """Calculates the system set using the user's currently selected profile"""
    system: Set[Atom] = set()
    for parent in profile_parents():
        packages = os.path.join(parent, "packages")
        if os.path.exists(packages):
            system |= {
                Atom(mod.lstrip("*"))
                for mod in read_list(packages)
                if mod.startswith("*")
            }

    return system



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3
"""Helper functions for getting win32-specific information"""

import sys


def get_personal() -> str:
    if sys.platform == "win32":
        import ctypes.wintypes

        CSIDL_PERSONAL = 5  # My Documents
        SHGFP_TYPE_CURRENT = 0  # Get current, not default value

        buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(
            None, CSIDL_PERSONAL, None, SHGFP_TYPE_CURRENT, buf
        )
        return buf.value
    raise Exception("get_personal should not be called on platforms other than win32!")



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Importing values from this module fills them with information about the
pybuild file from which they were imported
"""
# Note: This module should be removed from sys.modules prior to importing to ensure that
# the cached version is not used instead.

import os
from pathlib import Path
from typing import Optional

from portmod.atom import VAtom

CATEGORY: str
P: str
PF: str
PN: str
PV: str
PR: Optional[str]
PVR: str


def _set_info(filename):
    """Puts module information in the global scope so that it can be imported"""
    global P, PF, PN, PV, PR, PVR, CATEGORY
    if filename is not None:
        CATEGORY = Path(filename).resolve().parent.parent.name
        atom = VAtom(
            "{}/{}".format(CATEGORY, os.path.basename(filename)[: -len(".pybuild")])
        )

        P = atom.P
        PF = atom.PF
        PN = atom.PN
        PV = atom.PV
        PR = atom.PR
        PVR = atom.PV
        if atom.PR is not None:
            PVR += "-" + atom.PR



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Helper functions for interacting with the Windows registry

Also provides the following constants from the ``winreg`` builtin module::

    HKEY_CLASSES_ROOT: int
    HKEY_CURRENT_CONFIG: int
    HKEY_CURRENT_USER: int
    HKEY_LOCAL_MACHINE: int
    HKEY_USERS: int
"""

import sys

from portmod.winreg import read_reg

if sys.platform == "win32":
    from winreg import (  # noqa  # pylint: disable=unused-import,import-error,no-name-in-module
        HKEY_CLASSES_ROOT,
        HKEY_CURRENT_CONFIG,
        HKEY_CURRENT_USER,
        HKEY_LOCAL_MACHINE,
        HKEY_USERS,
    )

__all__ = [
    "read_reg",
]



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import re
from typing import AbstractSet, Any, Callable, List, Optional, Type

from portmod.atom import Atom, InvalidAtom, useflag_re
from portmod.l10n import l10n


def use_reduce(
    depstr: str,
    uselist: AbstractSet[str] = set(),
    masklist: AbstractSet[str] = set(),
    matchall: bool = False,
    excludeall: AbstractSet[str] = set(),
    is_src_uri: bool = False,
    opconvert: bool = False,
    flat: bool = False,
    is_valid_flag: Optional[Callable[[str], bool]] = None,
    token_class: Optional[Type] = None,
    matchnone: bool = False,
) -> List:
    """
    Takes a dep string and reduces the use? conditionals out, leaving an array
    with subarrays. All redundant brackets are removed.
    Adapted from portage's use_reduce

    args:
        depstr: depstring
        uselist: List of use enabled flags
        masklist: List of masked flags (always treated as disabled)
        param matchall: Treat all conditionals as active. Used by inquisitor.
        excludeall: List of flags for which negated conditionals are always treated
                    as inactive.
        is_src_uri: Indicates if depstr represents a SRC_URI
        opconvert: Put every operator as first element into it's argument list
        flat: Create a flat list of all tokens
        is_valid_flag: Function that decides if a given use flag might be used in
                       use conditionals
        token_class: Convert all non operator tokens into this class
        matchnone: Treat all conditionals as inactive. Used by digestgen().

    returns:
        The use reduced depend array
    """

    if opconvert and flat:
        raise ValueError(
            "portage.dep.use_reduce: 'opconvert' and 'flat' are mutually exclusive"
        )

    if matchall and matchnone:
        raise ValueError(
            "portage.dep.use_reduce: 'matchall' and 'matchnone' are mutually exclusive"
        )

    def is_active(conditional):
        """
        Decides if a given use conditional is active.
        """
        if conditional.startswith("!"):
            flag = conditional[1:-1]
            is_negated = True
        else:
            flag = conditional[:-1]
            is_negated = False

        if is_valid_flag:
            if not is_valid_flag(flag):
                msg = (
                    "USE flag '{}' referenced in conditional '{}' "
                    "is not in IUSE".format(flag, conditional)
                )
                raise Exception(msg)
        else:
            if useflag_re.match(flag) is None:
                raise Exception(
                    "invalid use flag '{}' in conditional '{}'".format(
                        flag, conditional
                    )
                )

        if is_negated and flag in excludeall:
            return False

        if flag in masklist:
            return is_negated

        if matchall:
            return True

        if matchnone:
            return False

        return (flag in uselist and not is_negated) or (
            flag not in uselist and is_negated
        )

    def missing_white_space_check(token, pos):
        """
        Used to generate good error messages for invalid tokens.
        """
        for x in (")", "(", "||"):
            if token.startswith(x) or token.endswith(x):
                raise Exception(
                    "missing whitespace around '%s' at '%s', token %s"
                    % (x, token, pos + 1)
                )

    mysplit = depstr.split()
    # Count the bracket level.
    level = 0
    # We parse into a stack.
    # Every time we hit a '(', a new empty list is appended to the stack.
    # When we hit a ')', the last list in the stack is merged with list one level up.
    stack: List[List[Any]] = [[]]
    # Set need_bracket to True after use conditionals or ||. Other tokens need to ensure
    # that need_bracket is not True.
    need_bracket = False
    # Set need_simple_token to True after a SRC_URI arrow. Other tokens need to ensure
    # that need_simple_token is not True.
    need_simple_token = False

    for pos, token in enumerate(mysplit):
        if token == "(":
            if need_simple_token:
                raise Exception(
                    "expected: file name, got: '%s', token %s" % (token, pos + 1)
                )
            if len(mysplit) >= pos + 2 and mysplit[pos + 1] == ")":
                raise Exception(
                    "expected: dependency string, got: ')', token %s" % (pos + 1,)
                )
            need_bracket = False
            stack.append([])
            level += 1
        elif token == ")":
            if need_bracket:
                raise Exception("expected: '(', got: '%s', token %s" % (token, pos + 1))
            if need_simple_token:
                raise Exception(
                    "expected: file name, got: '%s', token %s" % (token, pos + 1)
                )
            if level > 0:
                level -= 1
                ll = stack.pop()

                is_single = (
                    len(ll) == 1
                    or (opconvert and ll and ll[0] == "||")
                    or (not opconvert and len(ll) == 2 and ll[0] == "||")
                )
                ignore = False

                if flat:
                    # In 'flat' mode, we simply merge all lists into a single large one.
                    if stack[level] and stack[level][-1][-1] == "?":
                        # The last token before the '(' that matches the current ')'
                        # was a use conditional. The conditional is removed in any case.
                        # Merge the current list if needed.
                        if is_active(stack[level][-1]):
                            stack[level].pop()
                            stack[level].extend(ll)
                        else:
                            stack[level].pop()
                    else:
                        stack[level].extend(ll)
                    continue

                if stack[level] and isinstance(stack[level][-1], str):
                    if stack[level][-1] == "||" and not ll:
                        # Optimize: || ( ) -> .
                        ll.append((token_class or str)("empty-any-of"))
                        stack[level].pop()
                    elif stack[level][-1][-1] == "?":
                        # The last token before the '(' that matches the current ')'
                        # was a use conditional, remove it and decide if we
                        # have to keep the current list.
                        if not is_active(stack[level][-1]):
                            ignore = True
                        stack[level].pop()

                def ends_in_any_of_dep(k):
                    return k >= 0 and stack[k] and stack[k][-1] == "||"

                def last_any_of_operator_level(k):
                    # Returns the level of the last || operator if it is in effect for
                    # the current level. It is not in effect, if there is a level, that
                    # ends in a non-operator. This is almost equivalent to
                    # stack[level][-1]=="||", expect that it skips empty levels.
                    while k >= 0:
                        if stack[k] and isinstance(stack[k][-1], str):
                            if stack[k][-1] == "||":
                                return k
                            elif stack[k][-1][-1] != "?":
                                return -1
                        k -= 1
                    return -1

                def special_append():
                    """
                    Use extend instead of append if possible.
                    This kills all redundant brackets.
                    """
                    if is_single:
                        # Either [A], [[...]] or [|| [...]]
                        if ll[0] == "||" and ends_in_any_of_dep(level - 1):
                            if opconvert:
                                stack[level].extend(ll[1:])
                            else:
                                stack[level].extend(ll[1])
                        elif len(ll) == 1 and isinstance(ll[0], list):
                            # l = [[...]]
                            last = last_any_of_operator_level(level - 1)
                            if last == -1:
                                if (
                                    opconvert
                                    and isinstance(ll[0], list)
                                    and ll[0]
                                    and ll[0][0] == "||"
                                ):
                                    stack[level].append(ll[0])
                                else:
                                    stack[level].extend(ll[0])
                            else:
                                if opconvert and ll[0] and ll[0][0] == "||":
                                    stack[level].extend(ll[0][1:])
                                else:
                                    stack[level].append(ll[0])
                        else:
                            stack[level].extend(ll)
                    else:
                        if opconvert and stack[level] and stack[level][-1] == "||":
                            stack[level][-1] = ["||"] + ll
                        else:
                            stack[level].append(ll)

                if ll and not ignore:
                    # The current list is not empty and we don't want to ignore it
                    # because of an inactive use conditional.
                    if not ends_in_any_of_dep(level - 1) and not ends_in_any_of_dep(
                        level
                    ):
                        # Optimize: ( ( ... ) ) -> ( ... ).
                        # Make sure there is no '||' hanging around.
                        stack[level].extend(ll)
                    elif not stack[level]:
                        # An '||' in the level above forces us to keep to brackets.
                        special_append()
                    elif is_single and ends_in_any_of_dep(level):
                        # Optimize: || ( A ) -> A,  || ( || ( ... ) ) -> || ( ... )
                        stack[level].pop()
                        special_append()
                    elif ends_in_any_of_dep(level) and ends_in_any_of_dep(level - 1):
                        # Optimize: || ( A || ( B C ) ) -> || ( A B C )
                        stack[level].pop()
                        stack[level].extend(ll)
                    else:
                        if opconvert and ends_in_any_of_dep(level):
                            # In opconvert mode, we have to move the operator from the
                            # level above into the current list.
                            stack[level].pop()
                            stack[level].append(["||"] + ll)
                        else:
                            special_append()

            else:
                raise Exception(
                    "no matching '%s' for '%s', token %s" % ("(", ")", pos + 1)
                )
        elif token == "||":
            if is_src_uri:
                raise Exception(
                    "any-of dependencies are not allowed in SRC_URI: token %s"
                    % (pos + 1,)
                )
            if need_bracket:
                raise Exception("expected: '(', got: '%s', token %s" % (token, pos + 1))
            need_bracket = True
            stack[level].append(token)
        elif token == "->":
            if need_simple_token:
                raise Exception(
                    "expected: file name, got: '%s', token %s" % (token, pos + 1)
                )
            if not is_src_uri:
                raise Exception(
                    "SRC_URI arrow are only allowed in SRC_URI: token %s" % (pos + 1,)
                )
            need_simple_token = True
            stack[level].append(token)
        else:
            if need_bracket:
                raise Exception("expected: '(', got: '%s', token %s" % (token, pos + 1))

            if need_simple_token and "/" in token:
                # The last token was a SRC_URI arrow,
                # make sure we have a simple file name.
                raise Exception(
                    "expected: file name, got: '%s', token %s" % (token, pos + 1)
                )

            if token[-1] == "?":
                need_bracket = True
            else:
                need_simple_token = False
                if token_class and not is_src_uri:
                    # Add a hack for SRC_URI here, to avoid conditional code at the
                    # consumer level
                    try:
                        token = token_class(token)
                    except InvalidAtom as e:
                        missing_white_space_check(token, pos)
                        raise Exception("Invalid atom (%s), token %s" % (e, pos + 1))
                    except SystemExit:
                        raise
                    except Exception:
                        missing_white_space_check(token, pos)
                        raise Exception(
                            "Invalid token '%s', token %s" % (token, pos + 1)
                        )

                    if not matchall and isinstance(token, Atom):
                        token = token.evaluate_conditionals(uselist)

            stack[level].append(token)

    if level != 0:
        raise Exception("Missing '%s' at end of string" % (")",))

    if need_bracket:
        raise Exception("Missing '%s' at end of string" % ("(",))

    if need_simple_token:
        raise Exception("Missing file name at end of string")

    return stack[0]


def check_required_use(
    required_use: str, use: AbstractSet[str], iuse_match: Callable[[str], bool]
) -> bool:
    """
    Adapted from portage's check_required_use
    Checks if the use flags listed in 'use' satisfy all
    constraints specified in 'constraints'.

    args:
        required_use: REQUIRED_USE string
        use: Enabled use flags
        iuse_match: Callable that takes a single flag argument and returns
            True if the flag is matched, false otherwise,
    returns:
        Indicates if REQUIRED_USE constraints are satisfied
    """

    valid_operators = ("||", "^^", "??")

    def is_active(token):
        if token.startswith("!"):
            flag = token[1:]
            is_negated = True
        else:
            flag = token
            is_negated = False

        if not flag or not iuse_match(flag):
            raise Exception("USE flag '%s' is not in IUSE" % (flag,))

        return (flag in use and not is_negated) or (flag not in use and is_negated)

    def is_satisfied(operator, argument):
        if operator == "||":
            return True in argument
        elif operator == "^^":
            return argument.count(True) == 1
        elif operator == "??":
            return argument.count(True) <= 1
        elif operator[-1] == "?":
            return False not in argument

    mysplit = required_use.split()
    level = 0
    stack: List[List[str]] = [[]]
    need_bracket = False

    for token in mysplit:
        if token == "(":
            need_bracket = False
            stack.append([])
            level += 1
        elif token == ")":
            if need_bracket:
                raise Exception("malformed syntax: '%s'" % required_use)
            if level > 0:
                level -= 1
                ll = stack.pop()
                op = None
                if stack[level]:
                    if stack[level][-1] in valid_operators:
                        op = stack[level].pop()
                        satisfied = is_satisfied(op, ll)
                        stack[level].append(satisfied)

                    elif (
                        not isinstance(stack[level][-1], bool)
                        and stack[level][-1][-1] == "?"
                    ):
                        op = stack[level].pop()
                        if is_active(op[:-1]):
                            satisfied = is_satisfied(op, ll)
                            stack[level].append(satisfied)
                        else:
                            continue

                if op is None:
                    satisfied = False not in ll
                    if ll:
                        stack[level].append(satisfied)
            else:
                raise Exception("malformed syntax: '%s'" % required_use)
        elif token in valid_operators:
            if need_bracket:
                raise Exception("malformed syntax: '%s'" % required_use)
            need_bracket = True
            stack[level].append(token)
        else:
            if need_bracket:
                raise Exception("malformed syntax: '%s'" % required_use)

            if token[-1] == "?":
                need_bracket = True
                stack[level].append(token)
            else:
                satisfied = is_active(token)
                stack[level].append(satisfied)

    if level != 0 or need_bracket:
        raise Exception("malformed syntax: '%s'" % required_use)

    return False not in stack[0]


def parse_usestr(usestr: str, token_class: Type = str):
    """
    Adapted from portage's check_required_use
    Converts usestring into structure containing lists of tokens,
    possibly preceeded by an operator that can be one of || ?? ^^ or a use conditional
    """

    valid_operators = ("||", "^^", "??")

    mysplit = usestr.split()
    level = 0
    stack: List[List[Any]] = [[]]
    need_bracket = False

    for token in mysplit:
        if token == "(":
            need_bracket = False
            stack.append([])
            level += 1
        elif token == ")":
            if need_bracket:
                raise Exception(f"malformed syntax: '{usestr}'")
            if level > 0:
                level -= 1
                ll = stack.pop()
                stack[level].append(ll)
            else:
                raise Exception(f"malformed syntax: '{usestr}'")
        elif token in valid_operators:
            if need_bracket:
                raise Exception(f"malformed syntax: '{usestr}'")
            need_bracket = True
            stack[level].append(token)
        else:
            if need_bracket:
                raise Exception(f"malformed syntax: '{usestr}'")

            if token[-1] == "?":
                need_bracket = True
                stack[level].append(token)
            else:
                stack[level].append(token_class(token))

    if level != 0 or need_bracket:
        raise Exception(f"malformed syntax: '{usestr}'")

    def opconvert(tokenlist):
        index = 0
        while index < len(tokenlist):
            token = tokenlist[index]
            if (
                token in valid_operators
                or isinstance(token, str)
                and token.endswith("?")
            ):
                tokenlist[index] = [token] + opconvert(tokenlist[index + 1])
                del tokenlist[index + 1]
            index += 1
        return tokenlist

    # Nested ||, ^^, and ?? groups can be combined
    # use-conditionals cannot (tokens ending in ?)
    def flatten(parent):
        index = 0
        while index < len(parent):
            token = parent[index]
            if isinstance(token, list):
                token = flatten(token)
                if token[0] in valid_operators and token[0] == parent[0]:
                    parent += token[1:]
                    del parent[index]
                    index -= 1
                elif (
                    token[0] not in valid_operators
                    and parent[0] not in valid_operators
                    and not token[0].endswith("?")
                ):
                    parent += token
                    del parent[index]
                    index -= 1
            index += 1
        return parent

    return flatten(opconvert(stack[0]))


def human_readable_required_use(required_use: List[Any]) -> str:
    """Takes a (recursive) list of tokens and turns it into a readable string"""
    if isinstance(required_use[0], list):
        string = "( " + human_readable_required_use(required_use[0]) + " )"
    else:
        string = required_use[0]

    for token in required_use[1:]:
        if isinstance(token, list):
            string += " ( " + human_readable_required_use(token) + " )"
        else:
            string += " " + token
    return re.sub(
        r"\( (\!?\w+\?)",  # Move use-conditionals in front of brackets
        r"\1 (",
        string.replace("( ^^", l10n("exactly-one-of") + " (")
        .replace("( ||", l10n("any-of") + " (")
        .replace("( ??", l10n("at-most-one-of") + " ("),
    )



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the mod selection system
"""

import sys

import pytest

from portmod._cli.main import main
from portmod.atom import Atom
from portmod.loader import load_pkg
from portmod.util import LicenseDep, select_package

from .env import select_profile, setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup_repo():
    yield setup_env("test")
    tear_down_env()


@pytest.fixture()
def stable():
    """Sets up and cleans up the repo for tests"""
    select_profile("test")
    yield


@pytest.fixture()
def unstable():
    """Sets up and cleans up the repo for tests"""
    select_profile("test-unstable")
    yield


def test_select(stable):
    """Tests that we can select new mods over old ones"""
    mod1 = load_pkg(Atom("test/test-1.0"))[0]
    mod2 = load_pkg(Atom("test/test-2.0"))[0]
    assert select_package([mod1, mod2]) == (mod2, None)


def test_select_stable(stable):
    """
    Tests that we don't select unstable mods if a stable version is available
    and we only accept stable keywords
    """
    mod1 = load_pkg(Atom("test/test-1.0"))[0]
    mod2 = load_pkg(Atom("test/test-2.0"))[0]
    mod3 = load_pkg(Atom("test/test-2.0_rc1"))[0]
    assert select_package([mod1, mod2, mod3]) == (mod2, None)


def test_select_unstable(unstable):
    """
    Tests that we select unstable mods if available
    and we accept unstable keywords
    """
    mod1 = load_pkg(Atom("test/test-1.0"))[0]
    mod2 = load_pkg(Atom("test/test-2.0"))[0]
    mod3 = load_pkg(Atom("test/test-2.0_rc1"))[0]
    assert select_package([mod1, mod2, mod3]) == (mod2, None)


def test_select_eula(stable):
    """
    Tests that trying to select a mod with an EULA creates a LicenseDep
    with the eula flag enabled
    """
    mod1 = load_pkg(Atom("test/test-eula-1.0"))[0]
    mod, dep = select_package([mod1])
    assert mod == mod1
    assert isinstance(dep, LicenseDep)
    assert dep.is_eula


def test_select_revision(stable):
    """
    Tests that we can successfully select the mod with the highest revision
    """
    mod1 = load_pkg(Atom("test/test-1.0"))[0]
    mod2 = load_pkg(Atom("test/test-1.0-r1"))[0]
    mod3 = load_pkg(Atom("test/test-1.0-r2"))[0]
    assert select_package([mod1, mod2, mod3]) == (mod3, None)


def test_select_cli(stable):
    sys.argv = ["portmod", "test", "select", "profile", "list"]
    main()



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import stat
from functools import lru_cache
from shutil import Error, copy2, copystat
from typing import Callable, List, Optional, Set, Union

from portmod.portmod import _get_hash
from portmod.source import HashAlg

# 32MB buffer seems to give the best balance between performance on large files
# and on small files
HASH_BUF_SIZE = 32 * 1024 * 1024


def onerror(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise  # pylint: disable=misplaced-bare-raise


# Modified version of shutil.copytree from
# https://github.com/python/cpython
# Python software and documentation are licensed under the
# Python Software Foundation License Version 2
def patch_dir(
    src: Union[str, os.DirEntry],
    dst: str,
    *,
    symlinks: bool = False,
    ignore: Callable[[str, List[str]], Set[str]] = None,
    case_sensitive: bool = True,
) -> str:
    """
    Copies src ontop of dst

    args:
        src: Source directory to copy from
        dst: Destination directory to copy to
        symlinks: If true, copy symlinks as symlinks, if false, copy symlinks as the
            files they point to
        ignore: A callable which, given a directory and its contents, should return
            a set of files to ignore
        case_sensitive: If False, treat file and directory names as case insensitive
    """
    with os.scandir(src) as itr:
        entries = list(itr)
    if ignore is not None:
        ignored_names = ignore(os.fspath(src), [x.name for x in entries])
    else:
        ignored_names = set()

    if not os.path.isdir(dst):
        os.makedirs(dst)

    errors = []
    for entry in entries:
        if entry.name in ignored_names:
            continue
        srcname = os.path.join(src, entry.name)
        if case_sensitive:
            dstname = os.path.join(dst, entry.name)
        else:
            dstname = ci_exists(os.path.join(dst, entry.name)) or os.path.join(
                dst, entry.name
            )
        if os.path.exists(dstname) and entry.is_file():
            os.remove(dstname)
        try:
            is_symlink = entry.is_symlink()

            if is_symlink:
                linkto = os.readlink(srcname)
                if symlinks:
                    os.symlink(linkto, dstname)
                    copystat(entry, dstname, follow_symlinks=not symlinks)
                else:
                    if entry.is_dir():
                        patch_dir(
                            entry,
                            dstname,
                            symlinks=symlinks,
                            ignore=ignore,
                            case_sensitive=case_sensitive,
                        )
                    else:
                        copy2(entry, dstname)
            elif entry.is_dir():
                patch_dir(
                    srcname,
                    dstname,
                    symlinks=symlinks,
                    ignore=ignore,
                    case_sensitive=case_sensitive,
                )
            else:
                copy2(entry, dstname)
        except Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((srcname, dstname, str(why)))
    try:
        copystat(src, dst)
    except OSError as why:
        if getattr(why, "winerror", None) is None:
            errors.append((src, dst, str(why)))

    if errors:
        raise Error(errors)
    return dst


def ci_exists(path: str) -> Optional[str]:
    """
    Checks if a path exists, ignoring case.

    If the path exists but is ambiguous the result is not guaranteed
    """
    partial_path = "/"
    for component in os.path.normpath(os.path.abspath(path)).split(os.sep)[1:]:
        found = False
        for entryname in os.listdir(partial_path):
            if entryname.lower() == component.lower():
                partial_path = os.path.join(partial_path, entryname)
                found = True
                break
        if not found:
            return None

    if os.path.exists(partial_path):
        return partial_path

    return None


def get_tree_size(path):
    """Return total size of files in given path and subdirs."""
    total = 0
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            total += get_tree_size(entry.path)
        else:
            total += entry.stat(follow_symlinks=False).st_size
    return total


@lru_cache(maxsize=None)
def get_hash(filename: str, funcs=(HashAlg.BLAKE3,)) -> List[str]:
    """Hashes the given file"""
    return _get_hash(filename, [func.value for func in funcs], HASH_BUF_SIZE)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import json
import lzma
import os
import urllib
import urllib.parse
from functools import lru_cache
from typing import (
    AbstractSet,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from portmod.atom import Atom, FQAtom, QualifiedAtom
from portmod.globals import env
from portmod.parsers.manifest import Manifest
from portmod.parsers.usestr import check_required_use, use_reduce
from portmod.source import Source, SourceManifest, get_archive_basename


class File:
    """Represents important installed files and their metadata"""

    def __init__(
        self,
        NAME: str,
        REQUIRED_USE: str = "",
        OVERRIDES: Union[str, List[str]] = [],
        **kwargs,
    ):
        """
        File objects also support a REQUIRED_USE variable, for , and an OVERRIDES variable for overriding other plugins in the load order.
        """
        self.__keys__: Set[str] = set()
        self.NAME: str = NAME
        """Name of the file relative to the root of the InstallDir"""
        self.REQUIRED_USE: str = REQUIRED_USE
        """
        Requirements for installing this file

        The default empty string is always satisfied.
        See Pybuild1.REQUIRED_USE for details on the syntax.
        """
        self.OVERRIDES: Union[str, List[str]] = OVERRIDES
        """
        A list of files which this overrides when sorting (if applicable).

        Can either be in the form of a string containing use-conditionals (note that
        this does not support files that contain spaces) or a list of files to override.
        Note that these overridden files are not considered masters and do not need to
        be present.

        For archives it determines the order in which the fallback archives will be
        searched during VFS lookups.
        """
        if REQUIRED_USE:
            self._add_kwarg("REQUIRED_USE", REQUIRED_USE)
        if OVERRIDES:
            self._add_kwarg("OVERRIDES", OVERRIDES)

        for key in kwargs:
            self._add_kwarg(key, kwargs[key])

    def _add_kwarg(self, key, value):
        self.__dict__[key] = value
        self.__keys__.add(key)

    def __repr__(self):
        return self.__str__()

    def __str__(self) -> str:
        kvps = []
        for key in self.__keys__:
            value = getattr(self, key)
            if isinstance(value, str):
                kvps.append(f'{key}="{getattr(self, key)}"')
            else:
                kvps.append(f"{key}={getattr(self, key)}")

        separator = ""
        if kvps:
            separator = ", "
        return f'File("{self.NAME}"' + separator + ", ".join(kvps) + ")"

    def _to_cache(self):
        cache = {"NAME": self.NAME}
        for key in self.__keys__:
            cache[key] = getattr(self, key)

        cache["__type__"] = "File"
        return cache


class InstallDir:
    """
    Represents a directory in the Virtual File System

    Note that arbitrary arguments can be passed to the constructor, as
    repositories may make use of custom information.
    See the repository-level documentation for such information.
    """

    def __init__(
        self,
        PATH: str,
        REQUIRED_USE: str = "",
        PATCHDIR: str = ".",
        S: Optional[str] = None,
        WHITELIST: Optional[List[str]] = None,
        BLACKLIST: Optional[List[str]] = None,
        RENAME: Optional[str] = None,
        DATA_OVERRIDES: str = "",
        ARCHIVES: List[File] = [],
        **kwargs,
    ):
        self.PATH: str = PATH
        """
        The path to the data directory that this InstallDir represents
        relative to the root of the archive it is contained within.
        """
        self.REQUIRED_USE: str = REQUIRED_USE
        """
        A list of use flags with the same format as the package's
        REQUIRED_USE variable which enable the InstallDir if satisfied.
        Defaults to an empty string that is always satisfied.
        """
        self.PATCHDIR: str = PATCHDIR
        """
        The destination path of the InstallDir within the package's directory.

        Defaults to ".", i.e. the root of the mod directory. If multiple InstallDirs
        share the same PATCHDIR they will be installed into the same directory in the
        order that they are defined in the INSTALL_DIRS list.
        Each unique PATCHDIR has its own entry in the VFS, and its own sorting rules
        """
        self.S: Optional[str] = S
        """
        The source directory corresponding to this InstallDir.

        Similar function to S for the entire pybuild, this determines which directory
        contains this InstallDir, and generally corresponds to the name of the source
        archive, minus extensions. This is required for packages that contain more
        than one source, but is automatically detected for those with only one source
        if it is not specified, and will first take the value of Pybuild1.S, then the
        source's file name without extension if the former was not defined.
        """
        self.WHITELIST: Optional[List[str]] = WHITELIST
        """
        If present, only installs files matching the patterns in this list.
        fnmatch-style globbing patterns (e.g. * and [a-z]) can be used
        """
        self.BLACKLIST: Optional[List[str]] = BLACKLIST
        """
        If present, does not install files matching the patterns in this list.
        fnmatch-style globbing patterns (e.g. * and [a-z]) can be used
        """
        self.RENAME: Optional[str] = RENAME
        """
        Destination path of this directory within the final directory.

        E.g.::

            InstallDir("foo/bar", PATCHDIR=".", RENAME="bar")

        Will install the contents of ``foo/bar`` (in the source) into the directory
        ``bar`` inside the package's installation directory (and also the VFS).
        """
        self.DATA_OVERRIDES: str = DATA_OVERRIDES
        """
        A list of packages that this InstallDir should override in the VFS

        This only has a different effect from Pybuild1.DATA_OVERRIDES if multiple PATCHDIRs
        are set, as it can define overrides for individual PATCHDIRS, while
        Pybuild1.DATA_OVERRIDES affects all PATCHDIRs.
        See Pybuild1.DATA_OVERRIDES for details of the syntax.
        """
        self.ARCHIVES: List[File] = ARCHIVES
        """
        A list of File objects representing VFS archives.

        These will be searched, in order, during VFS file lookups if the file is not
        present in the package directories.
        """
        self.__keys__: Set[str] = set()
        if ARCHIVES:
            self._add_kwarg("ARCHIVES", ARCHIVES)
        if PATCHDIR != ".":
            self._add_kwarg("PATCHDIR", PATCHDIR)
        for arg in [
            "DATA_OVERRIDES",
            "RENAME",
            "BLACKLIST",
            "WHITELIST",
            "S",
            "REQUIRED_USE",
        ]:
            if getattr(self, arg):
                self._add_kwarg(arg, getattr(self, arg))
        for key in kwargs:
            self._add_kwarg(key, kwargs[key])

    def _add_kwarg(self, key, value):
        if isinstance(value, list):
            new_value = []
            for item in value:
                if isinstance(item, dict) and item.get("__type__") == "File":
                    file = dict(item)
                    file.pop("__type__", None)
                    new_value.append(File(**file))
                else:
                    new_value.append(item)
            value = new_value

        self.__dict__[key] = value
        self.__keys__.add(key)

    def get_files(self):
        """Generator function yielding file subattributes of the installdir"""
        for key in self.__dict__:
            if isinstance(getattr(self, key), list):
                for item in getattr(self, key):
                    if isinstance(item, File):
                        yield item

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        kvps = []
        for key in self.__keys__:
            value = getattr(self, key)
            if isinstance(value, str):
                kvps.append(f'{key}="{getattr(self, key)}"')
            else:
                kvps.append(f"{key}={getattr(self, key)}")

        separator = ""
        if kvps:
            separator = ", "
        return f'InstallDir("{self.PATH}"' + separator + ", ".join(kvps) + ")"

    def _to_cache(self):
        cache = {"PATH": self.PATH}
        for key in self.__keys__:
            value = getattr(self, key)
            if isinstance(value, list):
                new = []
                for item in value:
                    if isinstance(item, File):
                        new.append(item._to_cache())
                    else:
                        new.append(item)
                value = new
            cache[key] = value

        return cache


class Pybuild:
    """
    Interface describing the Pybuild Type
    Only describes elements that are cached.
    This class cannot be used to install/uninstall mods
    """

    __file__ = __file__

    ATOM: FQAtom
    RDEPEND: str = ""
    DEPEND: str = ""
    SRC_URI: str = ""
    P: Atom
    PF: Atom
    PN: Atom
    CATEGORY: str
    PV: str
    PR: str
    PVR: str
    CPN: QualifiedAtom
    CP: QualifiedAtom
    REQUIRED_USE: str = ""
    RESTRICT: str = ""
    PROPERTIES: str = ""
    IUSE_EFFECTIVE: Set[str] = set()
    IUSE: Set[str] = set()
    TEXTURE_SIZES: str = ""
    DESC: str = ""
    NAME: str = ""
    HOMEPAGE: str = ""
    LICENSE: str = ""
    KEYWORDS: str = ""
    REBUILD_FILES: List[str]
    TIER: str = "a"
    FILE: str
    REPO: str
    INSTALLED: bool = False
    INSTALL_DIRS: List[InstallDir]
    DATA_OVERRIDES = ""
    S: Optional[str]  # Primary directory during prepare and install operations:w
    PATCHES: str = ""
    # Phase functions defined by the pybuild (or superclasses other than Pybuild1)
    # Used to determine if a function should be run, as certain functions don't have any default
    # behaviour
    FUNCTIONS: List[str] = []

    def __init__(self, atom: FQAtom, cache: Optional[Dict] = None, **kwargs):
        # Note: mypy doesn't like how we coerce INSTALL_DIRS
        if cache:
            self.__dict__ = cache
            self.INSTALL_DIRS = [
                InstallDir(**cast(Dict, idir)) for idir in self.INSTALL_DIRS
            ]
        for keyword, value in kwargs.items():
            setattr(self, keyword, value)
        self.ATOM = atom
        self.P = Atom(atom.P)
        self.PF = Atom(atom.PF)
        self.PN = Atom(atom.PN)
        self.CATEGORY = atom.C
        self.PV = atom.PV
        self.PR = atom.PR or "r0"
        self.PVR = atom.PVR
        self.CPN = QualifiedAtom(atom.CPN)
        self.CP = QualifiedAtom(atom.CP)
        self.__ENV = None
        self.INSTALLED = False

    def valid_use(self, use: str) -> bool:
        """Returns true if the given flag is a valid use flag for this mod"""
        return use in self.IUSE_EFFECTIVE

    @lru_cache()
    def get_installed_env(self):
        """Returns a dictionary containing installed object values"""
        if not self.INSTALLED:
            raise Exception("Trying to get environment for mod that is not installed")

        path = os.path.join(os.path.dirname(self.FILE), "environment.xz")
        if os.path.exists(path):
            environment = lzma.LZMAFile(path)
            try:
                return json.load(environment)
            except EOFError as e:
                raise RuntimeError(f"Failed to read {path}") from e

        return {}

    def get_dir_path(self, install_dir: InstallDir) -> str:
        """Returns the installed path of the given InstallDir"""
        path = os.path.normpath(
            os.path.join(
                env.prefix().PACKAGE_DIR, self.CATEGORY, self.PN, install_dir.PATCHDIR
            )
        )
        if os.path.islink(path):
            return os.readlink(path)
        else:
            return path

    def get_file_path(self, install_dir: InstallDir, esp: File) -> str:
        return os.path.join(self.get_dir_path(install_dir), esp.NAME)

    def __str__(self):
        return self.ATOM

    def __repr__(self):
        return self.__class__.__name__ + "(" + self.FILE + ")"

    def get_manifest(self):
        """Returns the manifest object for the mod's sources"""
        if os.path.exists(manifest_path(self.FILE)):
            self._manifest = get_manifest(self.FILE)

        if hasattr(self, "_manifest"):
            return self._manifest

        return get_manifest(self.FILE)

    def _get_sources(
        self,
        uselist: AbstractSet[str] = frozenset(),
        masklist: AbstractSet[str] = frozenset(),
        matchnone=False,
        matchall=False,
    ) -> List[Source]:
        """
        Returns a list of sources that are enabled using the given configuration
        """
        sourcestr = self.SRC_URI
        sources = use_reduce(
            sourcestr,
            uselist,
            masklist,
            is_valid_flag=self.valid_use,
            is_src_uri=True,
            flat=True,
            matchnone=matchnone,
            matchall=matchall,
        )
        return parse_arrow(sources)

    def get_default_sources(self) -> List[SourceManifest]:
        """
        Returns a list of sources that are enabled
        with the current use configuration
        """
        return self.get_sources(*self.get_use())

    def get_sources(
        self,
        uselist: AbstractSet[str] = frozenset(),
        masklist: AbstractSet[str] = frozenset(),
        matchnone=False,
        matchall=False,
    ) -> List[SourceManifest]:
        """
        Returns a list of sources that are enabled using the given configuration
        """
        sources = self._get_sources(uselist, masklist, matchnone, matchall)
        manifest = self.get_manifest()

        manifested_sources: List[SourceManifest] = []

        for source in sources:
            if manifest.get(source.name) is not None:
                m = manifest.get(source.name)
                manifested_sources.append(SourceManifest(source, m.hashes, m.size))
            else:
                raise Exception(f"Source {source.name}  is missing from the manifest!")

        return manifested_sources

    def get_use(self) -> Tuple[Set[str], Set[str]]:
        """Returns the use flag configuration for the mod"""
        from .config.use import get_use

        return get_use(self)

    def get_enabled_use(self) -> Set[str]:
        """Returns the enabled use flags for the mod"""
        # FIXME: This should become the default behaviour of get_use
        # For the few cases where we need explicitly disabled flags,
        # we can use a different function
        return self.get_use()[0]

    def parse_string(self, string, matchall=False):
        if not matchall:
            (enabled, disabled) = self.get_use()
        else:
            (enabled, disabled) = (set(), set())

        return use_reduce(
            self.RESTRICT,
            enabled,
            disabled,
            is_valid_flag=self.valid_use,
            flat=True,
            matchall=matchall,
        )

    def get_restrict(self, *, matchall=False):
        """Returns parsed tokens in RESTRICT using current use flags"""
        # If we don't have a prefix there is no user configuration
        if not env.PREFIX_NAME:
            matchall = True
        return self.parse_string(self.RESTRICT, matchall=matchall)

    def get_properties(self, *, matchall=False):
        """Returns parsed tokens in PROPERTIES using current use flags"""
        return self.parse_string(self.PROPERTIES, matchall=matchall)

    def get_directories(self) -> Generator[InstallDir, None, None]:
        """
        Returns all enabled InstallDir objects in INSTALL_DIRS
        """
        for install_dir in self.INSTALL_DIRS:
            if check_required_use(
                install_dir.REQUIRED_USE, self.get_enabled_use(), self.valid_use
            ):
                yield install_dir

    def get_files(self, typ: str) -> Generator[Tuple[InstallDir, File], None, None]:
        """
        Returns all enabled files and their directories
        """
        for install_dir in self.get_directories():
            if hasattr(install_dir, typ):
                for file in getattr(install_dir, typ):
                    if check_required_use(
                        file.REQUIRED_USE, self.get_enabled_use(), self.valid_use
                    ):
                        yield install_dir, file

    def get_default_source_basename(self) -> Optional[str]:
        tmp_source = next(iter(self.get_default_sources()), None)
        if tmp_source:
            return get_archive_basename(tmp_source.name)
        return None


# Class used for typing pybuilds, allowing more flexibility with
# the implementations. Implementations of this class (e.g. Pybuild1)
# should derive it, but build file Mod classes should derive one of
# the implementations. This should be used as the type for any function that
# handles Pybuild objects.
#
# This provides a mechanism for modifying the Pybuild format, as we can
# make changes to this interface, and update the implementations to conform
# to it while keeping their file structure the same, performing conversions
# of the data inside the init function.
class FullPybuild(Pybuild):
    """Interface describing the Pybuild Type"""

    __file__ = __file__
    TIER: str
    REPO_PATH: str
    __pybuild__: str

    # Variables defined during the install/removal process
    A: List[Source]  # List of enabled sources
    D: str  # Destination directory where the mod is to be installed
    FILESDIR: str  # Path of the directory containing additional repository files
    ROOT: str  # Path of the installed directory of the mod
    T: str  # Path of temp directory
    UNFETCHED: List[Source]  # List of sources that need to be fetched
    USE: Set[str]  # Enabled use flags
    WORKDIR: str  # Path of the working directory

    def pkg_nofetch(self):
        """
        Function to give user instructions on how to fetch a mod
        which cannot be fetched automatically
        """

    def pkg_pretend(self):
        """
        May be used to carry out sanity checks early on in the install process

        Note that the default does nothing, and it will not even be executed unless defined.
        """

    def src_unpack(self):
        """Function used to unpack mod sources"""

    def src_prepare(self):
        """Function used to apply patches and configuration"""

    def src_install(self):
        """Function used to create the final install directory"""

    def pkg_prerm(self):
        """
        Function called immediately before mod removal

        Note that the default does nothing, and it will not even be executed unless defined.
        """

    def pkg_postinst(self):
        """
        Function called immediately after mod installation

        Note that the default does nothing, and it will not even be executed unless defined.
        """

    def _validate(self):
        """QA Checks pybuild structure"""

    @staticmethod
    def execute(
        command: str, pipe_output: bool = False, pipe_error: bool = False
    ) -> Optional[str]:
        """Function pybuild files can use to execute native commands"""

    def _to_cache(self) -> Dict:
        from pybuild import Pybuild1

        cache = {}
        for key in [
            "RDEPEND",
            "DEPEND",
            "SRC_URI",
            "REQUIRED_USE",
            "RESTRICT",
            "PROPERTIES",
            "IUSE_EFFECTIVE",
            "IUSE",
            "TEXTURE_SIZES",
            "DESC",
            "NAME",
            "HOMEPAGE",
            "LICENSE",
            "KEYWORDS",
            "REBUILD_FILES",
            "TIER",
            "FILE",
            "REPO",
            "DATA_OVERRIDES",
            "S",
            "PATCHES",
        ]:
            cache[key] = getattr(self, key)

        cache["INSTALL_DIRS"] = [idir._to_cache() for idir in self.INSTALL_DIRS]
        phase_functions = [
            "src_unpack",
            "src_install",
            "src_prepare",
            "pkg_nofetch",
            "pkg_pretend",
            "pkg_postinst",
            "pkg_prerm",
        ]
        cache["FUNCTIONS"] = [
            func
            for func in phase_functions
            if getattr(self.__class__, func) != getattr(Pybuild1, func)
        ]
        return cache


class InstalledPybuild(Pybuild):
    """Interface describing the type of installed Pybuilds"""

    INSTALLED_USE: Set[str] = set()
    INSTALLED_REBUILD_FILES: Optional[Manifest] = None

    def __init__(self, atom: FQAtom, cache: Optional[Dict] = None, **kwargs):
        super().__init__(atom, cache=cache, **kwargs)
        self.INSTALLED_USE = set(self.INSTALLED_USE)
        self.INSTALLED = True
        if self.INSTALLED_REBUILD_FILES:
            self.INSTALLED_REBUILD_FILES = Manifest.from_json(
                self.INSTALLED_REBUILD_FILES
            )

    def get_enabled_use(self):
        return self.INSTALLED_USE


class FullInstalledPybuild(InstalledPybuild, FullPybuild):
    """Interface describing the type of installed Pybuilds"""

    INSTALLED_USE: Set[str]

    def get_installed_env(self):
        """Returns a dictionary containing installed object values"""

    def _to_cache(self) -> Dict:
        cache = FullPybuild._to_cache(self)
        cache["INSTALLED_USE"] = self.INSTALLED_USE
        cache["INSTALLED_REBUILD_FILES"] = None
        if self.INSTALLED_REBUILD_FILES:
            cache["INSTALLED_REBUILD_FILES"] = self.INSTALLED_REBUILD_FILES.to_json()
        return cache


def parse_arrow(sourcelist: Iterable[str]) -> List[Source]:
    """
    Turns a list of urls using arrow notation into a list of
    Source objects
    """
    result: List[Source] = []
    arrow = False
    for value in sourcelist:
        if arrow:
            result[-1] = Source(result[-1].url, value)
            arrow = False
        elif value == "->":
            arrow = True
        else:
            url = urllib.parse.urlparse(value)
            result.append(Source(value, os.path.basename(url.path)))
    return result


def manifest_path(file):
    return os.path.join(os.path.dirname(file), "Manifest")


# Loads the manifest for the given file, i.e. the Manifest file in the same directory
#    and turns it into a map of filenames to (shasum, size) pairs
def get_manifest(file):
    m_path = manifest_path(file)

    return Manifest(m_path)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import re
from functools import lru_cache
from logging import info
from typing import Iterable, Optional, Set

from portmod.atom import Atom, atom_sat
from portmod.l10n import l10n

from .list import read_list


def get_flags(file: str, atom: Optional[Atom] = None) -> Set[str]:
    """
    Reads flags from a given file.
    If the atom argument is passed, file is assumed to be newline delimited
    with an atom followed by a list of flags on each line.
    If atom is None, file is assumed to be a newline delimited list of flags
    """
    flags: Set[str] = set()
    if os.path.exists(file):
        for line in read_list(file):
            # Remove comments
            line = re.sub("#.*", "", line)
            if not line:
                continue

            if atom:
                elem = line.split()
                line_atom = Atom(elem[0])
                if atom_sat(atom, line_atom):
                    flags = collapse_flags(flags, elem[1:])
            else:
                flags = collapse_flags(flags, line)
        return flags
    return set()


def collapse_flags(old: Iterable[str], new: Iterable[str]) -> Set[str]:
    """
    Collases an ordered list of flags into an unordered set of flags
    """
    newset = set(old)
    for flag in new:
        if not flag.startswith("-"):
            newset.discard(f"-{flag}")
        elif flag.startswith("-"):
            newset.discard(flag.lstrip("-"))
        newset.add(flag)
    return newset


def add_flag(file, atom, flag):
    if os.path.exists(file):
        flagfile = __read_flags(file)
    else:
        flagfile = []

    found = False
    for (index, line) in enumerate(flagfile):
        tokens = line.split()
        if tokens[0] == atom:
            if flag not in tokens:
                info(l10n("flag-add", flag=flag, atom=atom, file=file))
                flagfile[index] = "{} {}".format(line, flag)
            found = True
    if not found:
        info(l10n("flag-add", flag=flag, atom=atom, file=file))
        flagfile.append("{} {}".format(atom, flag))

    __write_flags(file, flagfile)


def remove_flag(file, atom, flag):
    flagfile = __read_flags(file)

    for (index, line) in enumerate(flagfile):
        tokens = line.split()
        if atom_sat(Atom(tokens[0]), atom) and flag in tokens:
            info(l10n("flag-remove", flag=flag, atom=atom, file=file))
            tokens = list(filter(lambda a: a != flag, tokens))

            if len(tokens) > 1:
                flagfile[index] = " ".join(tokens)
            else:
                del flagfile[index]

    __write_flags(file, flagfile)


@lru_cache()
def __read_flags(file):
    if os.path.exists(file):
        with open(file, mode="r") as flagfile:
            return flagfile.read().splitlines()
    return []


def __write_flags(file, new_flagfile):
    with open(file, mode="w") as flagfile:
        for line in new_flagfile:
            print(line, file=flagfile)
    __read_flags.cache_clear()
